/*
 *  Common Services Library - Framework plug-in.
 *  (C) Copyright Fundi Software 2011, 2016.  Fundi Confidential.
 *  All rights reserved.  Use, duplication or disclosure restricted. 
 */

package com.fundi.framework.common.change;

import java.util.HashMap;
import java.util.LinkedHashMap;
import java.util.Map;

public class ChangeListScope {

	//Eye catchers.
	public static final String EC_COPYRIGHT = "(C) Copyright 2011, 2016 Fundi Software.  Fundi Confidential."; //$NON-NLS-1$
	public static final String EC_APAR = "3.1.1.00_00014"; //$NON-NLS-1$
	public static final String EC_FMID = "H30SA10"; //$NON-NLS-1$
	public static final String EC_PID = "5655-CSL"; //$NON-NLS-1$
	
	/**
	 * The intention here is to introduce a change-tracking model to the GUI without over-stepping the scope of current development.
	 * ChangeListScope has two orthogonal fields:
	 * 		changeItems - to track the set of objects over which all changes are to be applied.
	 * 		changeList	- to track the list of changes to be applied to the set of changeItems.
	 * 
	 * Because of the desire to limit the scope, changes are not tracked against individual changeItems, and repeated changes to the 
	 * one property are not incrementally stored. However, it is expected that these may be incorporated into the existing structure 
	 * at some point, probably by producing new subclasses, so the design and naming patterns attempt to take these features into account.
	 * 
	 * This implementation will only accept changes for existing fields, anything else is ignored (null is returned as the ChangeDetail).
	 */
	
	public enum OperationType {
		CREATE,
		UPDATE,
		DELETE;
		//don't need READ in here - implicitly there are no changes made on READ
	}
	
	private IChangeable						defaultItem;
	private Map<Object, IChangeable>		changeItems;
	//private List<IChange>					changeList;		//allows recording of an absolute order of all items and details - but at this stage, unimplemented
	private Map<Object, IChange>			changeListIndex;	//always holds the reference to the latest change object for each key - allows very quick getinitial/getfinal operations

//Constructors
	public ChangeListScope() {
		changeListIndex = new LinkedHashMap<Object, IChange>();
		changeItems = new HashMap<Object, IChangeable>();
	}

//API methods
	public IChangeable add(IChangeable changeable) {
		IChangeable	changeItem;
		
		if (changeable.getChangeListScope() != null) {
			changeable.getChangeListScope().undo(changeable.getSource());
		}
		changeable.setChangeListScope(this);
		changeItem = get(changeable.getSource());
		if (changeItem == null) {
			changeItems.put(changeable.getSource(), changeable);
		}
		
		return changeItem;
	}

	public IChange setValue(Object key, Object value) {
		ChangeDetail	changeDetail;
		
		changeDetail = null;
		if (getDefaultItem().isRecognisedKey(key)) {
			changeDetail = new ChangeDetail(key, value);
			changeListIndex.remove(key);
			if (isChange(key, value)) {
				changeListIndex.put(key, changeDetail);
			}
		}
		
		return changeDetail;
	}

	public IChangeable get(Object source) {	//may return null
		return changeItems.get(source);
	}

	public Object getValue(Object key) {
		Object	result;
		IChange change;
		
		result = null;
		change = null;
		if (getDefaultItem().isRecognisedKey(key)) {			//allow default item to control visibility of current set of changes
			change = changeListIndex.get(key);
		}
		if (change != null) {
			result = change.getValue();
		}
		
		return result;
	}
	
	public Map<Object, IChange> getChangeDetails() {
		Map<Object, IChange> changes;
		
		changes = new LinkedHashMap<Object, IChange>();
		for (IChange change: changeListIndex.values()) {
			if (getDefaultItem().isRecognisedKey(change.getKey())) {			//allow default item to control visibility of current set of changes
				changes.put(change.getKey(), change);
			}
		}
		
		return changes;
	}
	
	public Map<Object, Object> getValues() {
		Map<Object, Object> changes;
		
		changes = new LinkedHashMap<Object, Object>();
		for (IChange change: changeListIndex.values()) {
			if (getDefaultItem().isRecognisedKey(change.getKey())) {			//allow default item to control visibility of current set of changes
				changes.put(change.getKey(), change.getValue());
			}
		}
		
		return changes;
	}
	
	public void clear() {
		clearItems();
		clearChanges();
	}
	
	public void clearChanges() {
		changeListIndex.clear();
	}
	
	public void clearItems() {
		changeItems.clear();
	}
	
	public void undo(Object source) {	
		changeItems.remove(source);
	}

	public void undoChange(Object key) {
		changeListIndex.remove(key);
	}
	
	public IChangeable getReferenceItem() {
		IChangeable changeable;
		
		changeable = defaultItem;
		if (changeItems.size() == 1) {
			changeable = changeItems.values().iterator().next();
		}
		
		return changeable;
	}
	
	public boolean isInitiallySpecified(Object key) {
		return isInitiallySpecified(getReferenceItem(), key);
	}

	public boolean isInitiallySpecified(Object source, Object key) {
		return isInitiallySpecified(get(source), key);
	}

	public boolean isInitiallySpecified(IChangeable source, Object key) {
		boolean result = false;
		Object value;
		
		if (source != null) {
			value = source.getInitialValue(key);
			result = value != null;
			if (result) {
				if (String.class.isInstance(value)) {
					result = !((String) value).isEmpty();
				}
			}
		}
		
		return result;
	}
	
	public Object getDefaultValue(Object key) {
		Object		result;
		
		result = null;
		if (defaultItem != null) {
			result = defaultItem.getInitialValue(key);	//changeable object is expected to handle returning of defaults if appropriate;
		}
		
		return result;
	}
	
	public Object getInitialValue(Object source, Object key) {
		IChangeable changeable;
		Object		result;
		
		result = null;
		changeable = changeItems.get(source); 
		if (changeable == null) {
			changeable = getDefaultItem();	//could theoretically be retrieving the same item
		}
		if (changeable != null) {
			result = changeable.getInitialValue(key);
		}
		
		return result;
	}
	
	public Object getFinalValue(Object source, Object key) {
		Object		result;
		
		result = getValue(key);
		if (result == null) {
			result = getInitialValue(source, key);
		}
		
		return result;
	}

	public IChangeable getDefaultItem() {
		return defaultItem;
	}

	public void setDefaultItem(IChangeable defaultItem) {
		this.defaultItem = defaultItem;
	}

	public boolean isChanged() {
		return !changeListIndex.isEmpty();
	}
	
	private boolean isChange(Object key, Object value) {
		Object refValue;
		
		refValue = getReferenceItem().getInitialValue(key);
		if (refValue == null) {
			refValue = getDefaultValue(key);
		}
		return !isNullOrEqual(value, refValue); 
	}
	
	public boolean isNullOrEqual(Object o1, Object o2) {
		boolean result;
		
		if (o1 == null || o2 == null) {
			result = o1 == null && o2 == null;
		} else {
			result = o1.equals(o2);
		}
		
		return result;
	}

	public int itemCount() {
		return changeItems.size();
	}

	public Map<Object, IChangeable> getChangeItems() {
		return changeItems;
	}

	@Override
	public String toString() {
		StringBuilder	result = new StringBuilder();
		IChangeable		referenceItem,
						defaultItem;
		
		result.append("ChangeListScope:"); //$NON-NLS-1$
		referenceItem = getReferenceItem();
		result.append(" ReferenceItem = "); //$NON-NLS-1$
		result.append(referenceItem != null ? referenceItem.toString() : "<null>");	 //$NON-NLS-1$
		defaultItem = getDefaultItem();
		result.append(" DefaultItem = "); //$NON-NLS-1$
		result.append(defaultItem != null ? defaultItem.toString() : "<null>"); //$NON-NLS-1$
		result.append(" Item count = "); //$NON-NLS-1$
		result.append(changeItems.size());
		result.append(" Change count = "); //$NON-NLS-1$
		result.append(changeListIndex.size());
		
		return result.toString();
	}
	
}
