package com.fundi.framework.common.instance;

/*
 *  Common Services Library - Framework plug-in.
 *  (C) Copyright Fundi Software 2011, 2016.  Fundi Confidential.
 *  All rights reserved.  Use, duplication or disclosure restricted. 
 */

import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.FileWriter;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.lang.reflect.Method;
import java.net.URL;
import java.security.Key;
import java.security.SecureRandom;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Date;
import java.util.Dictionary;
import java.util.Iterator;
import java.util.List;
import java.util.concurrent.Executors;
import java.util.concurrent.ScheduledExecutorService;
import java.util.concurrent.ScheduledFuture;
import java.util.concurrent.TimeUnit;

import javax.crypto.Cipher;
import javax.crypto.KeyGenerator;
import javax.xml.bind.annotation.XmlRootElement;

import org.eclipse.core.runtime.IConfigurationElement;
import org.eclipse.core.runtime.IExtensionRegistry;
import org.eclipse.core.runtime.Platform;
import org.osgi.framework.BundleContext;

import com.fundi.framework.common.comparator.ComparatorHelper;
import com.fundi.framework.common.comparator.ComparatorSource;
import com.fundi.framework.common.displaylist.DisplayList;
import com.fundi.framework.common.displaylist.DisplayListDeserializer;
import com.fundi.framework.common.displaylist.ListFieldColor;
import com.fundi.framework.common.displaylist.ListLayoutV1;
import com.fundi.framework.common.editor.ISavedDisplayManager;
import com.fundi.framework.common.editor.ObjectEditor;
import com.fundi.framework.common.editor.SavedDisplay;
import com.fundi.framework.common.helpers.IListLayoutHelper;
import com.fundi.framework.common.helpers.ListFilterHelper;
import com.fundi.framework.common.helpers.ListLayoutHelper;
import com.fundi.framework.common.navigation.FrameworkNavigationProvider;
import com.fundi.framework.common.navigation.INavigationType;
import com.fundi.framework.common.navigation.MenuItem;
import com.fundi.framework.common.navigation.MenuItem.Style;
import com.fundi.framework.common.navigation.IEditContext;
import com.fundi.framework.common.navigation.NavigationItem;
import com.fundi.framework.common.navigation.NavigationProvider;
import com.fundi.framework.common.navigation.NavigationType;
import com.fundi.framework.common.navigation.NavigationItemEditor;
import com.fundi.framework.common.navigation.SerializableNavigationItem;
import com.fundi.framework.common.nl1.Messages;
import com.fundi.framework.common.preferences.PreferenceCategory;
import com.fundi.framework.common.preferences.PreferenceEditor;
import com.fundi.framework.common.util.NumberEditor;
import com.fundi.framework.common.util.StringEditor;

import sun.misc.BASE64Decoder;
import sun.misc.BASE64Encoder;

@XmlRootElement
public class AppInstance {

	//Eye catchers.
	public static final String EC_COPYRIGHT = "(C) Copyright 2011, 2016 Fundi Software.  Fundi Confidential."; //$NON-NLS-1$
	public static final String EC_APAR = "3.1.1.00_00014"; //$NON-NLS-1$
	public static final String EC_FMID = "H30SA10"; //$NON-NLS-1$
	public static final String EC_PID = "5655-CSL"; //$NON-NLS-1$
	
	/**
	 *  This class represents an application instance.
	 * 
	 *  It contains functionality common to all instance types (e.g. Eclipse, Dojo), such as encoding/decoding, reading parameter files, managing
	 *  user login/switching, user display preferences. 
	 *  
	 *  When an application loads, it should create an AppInstance.
	 *  
	 *  In an Eclipse environment userFolder is not required at initialisation because in Eclipse the working folder should be in the workspace and
	 *  hence not shared between users - pass an empty string.  In Dojo the context may apply to many users, so the working folder needs to
	 *  be user specific - pass a value for userFolder so that the Framework knows where to store persistent information across sessions.  
	 *  
	 *  Note that AppInstance will validate the user if at previous shutdown rememberConnectionSettings was true.  In an Eclipse environment
	 *  the userid is not required to do this, but in a Dojo environment it is required - this is because in Eclipse the working folder should be in the
	 *  workspace and hence not shared between users, but in Dojo the context may apply to many users.  
	 *  
	 *  If running under IBM's DAI architecture, it is expected the userid will have been previously obtained by a DAI-managed login process,
	 *  and hence is known during initialisation of AppInstance.
	 */

	private static KeyGenerator keyGenerator;
	private static Cipher cipher = null;
	private static Key secretKey = null;
	private static final String cipherAlgorithm = Messages.getString("AppInstance_4"); //$NON-NLS-1$
	
	public static final String instanceTypeDojo = Messages.getString("AppInstance_5"); //$NON-NLS-1$
	public static final String instanceTypeEclipse = Messages.getString("AppInstance_6"); //$NON-NLS-1$
	
	public static final String timezoneGMT = Messages.getString("AppInstance_7"); //$NON-NLS-1$
	public static final String timezoneServer = Messages.getString("AppInstance_8"); //$NON-NLS-1$
	public static final String timezoneClient = Messages.getString("AppInstance_9"); //$NON-NLS-1$
	
	public static final String messageOK = ""; //$NON-NLS-1$
	public static final String messageNoUser = Messages.getString("AppInstance_11"); //$NON-NLS-1$
	public static final String messageInvalidUserPasswordCombination = Messages.getString("AppInstance_12"); //$NON-NLS-1$
	public static final String messageNoPassword = Messages.getString("AppInstance_13"); //$NON-NLS-1$
	public static final String messageOldPasswordRequired = Messages.getString("AppInstance_14"); //$NON-NLS-1$
	public static final String messageNoTimezone = Messages.getString("AppInstance_15"); //$NON-NLS-1$
	public static final String messageInvalidTestInterval = Messages.getString("AppInstance_16"); //$NON-NLS-1$
	
	private static final String instanceFilename = Messages.getString("AppInstance_17"); //$NON-NLS-1$
	private String instanceType = null;
	
	public static final int defaultSystemsRefreshInterval = 5; //Minutes.
	
	private SerializableAppInstance instance;
	
	public static final String fudProductId = "FUD"; //$NON-NLS-1$
	
	private File workingFolder;
	
	private ArrayList<NavigationProvider> navigationProviders = new ArrayList<NavigationProvider>();
	private ArrayList<ObjectEditor> objectEditors = new ArrayList<ObjectEditor>();
	
	public static final String userGroupType = Messages.getString("AppInstance_18"); //$NON-NLS-1$
	public static final String allSystemsGroupCaption = Messages.getString("AppInstance_19"); //$NON-NLS-1$
	public static final String connectionServerGroupCaption = Messages.getString("AppInstance_20"); //$NON-NLS-1$
	public static final String allNavigationItemTypes = Messages.getString("AppInstance_21"); //$NON-NLS-1$
	private static final String connectionServerFilenamePfx = Messages.getString("AppInstance_22"); //$NON-NLS-1$
	private static final String openItemsFilename = Messages.getString("AppInstance_42"); //$NON-NLS-1$
	private static final String expandedNavItemsFilename = Messages.getString("AppInstance_67"); //$NON-NLS-1$
	private static final String connectionProfilesFilename = Messages.getString("AppInstance_68"); //$NON-NLS-1$
	private static final String defaultConnectionProfile = Messages.getString("AppInstance_69");  //$NON-NLS-1$
	private static final String listLayoutsFilename = Messages.getString("AppInstance_71"); //$NON-NLS-1$
	
	private String filterType = "";
	private String searchValue = "";
	private ArrayList<NavigationItem> allGroupedItems = new ArrayList<NavigationItem>();
	private ArrayList<NavigationItem> allItems = new ArrayList<NavigationItem>();
	private ArrayList<NavigationItem> viewableItems = new ArrayList<NavigationItem>();
	private ArrayList<NavigationItem> connectionServerItems = new ArrayList<NavigationItem>();
	private ArrayList<NavigationRebuildListener> navigationRebuildListeners = new ArrayList<NavigationRebuildListener>();
	private ArrayList<ConnectionProfileValidationListener> connectionProfileEditListeners = new ArrayList<ConnectionProfileValidationListener>();
	private ArrayList<NavigationFocusListener> navigationFocusListeners = new ArrayList<NavigationFocusListener>();
	
	private ArrayList<ConnectionServer> connectionServers = new ArrayList<ConnectionServer>();
	
	private ArrayList<ConsoleMessage> consoleMessages = new ArrayList<ConsoleMessage>(); //$NON-NLS-1$
	private ArrayList<ConsoleListener> consoleListeners = new ArrayList<ConsoleListener>();
	
	private ArrayList<String> loadedBundleNames = new ArrayList<String>();
	private ArrayList<String> loadedBundleVersions = new ArrayList<String>();
	
	public static final int CAN_CONNECT_UNKNOWN = 0;
	public static final int CAN_CONNECT_OK = 1;
	public static final int CAN_CONNECT_UNABLE = 2;
	public static final int CAN_CONNECT_AUTHENTICATION_PROBLEM = 3;
	
	private ScheduledExecutorService systemRefreshService = Executors.newScheduledThreadPool(1);
	private ScheduledFuture<?> refreshScheduler = null;
	
	private ArrayList<ComparatorSource> openComparatorSources = new ArrayList<ComparatorSource>();
	private ArrayList<ComparatorSource> savedDisplayComparatorSources = new ArrayList<ComparatorSource>();
	private int highestComparatorSourceId = -1;
	public static final String savedDisplayPfx = Messages.getString("AppInstance_0"); //$NON-NLS-1$  //Only used for old-style saved displays.
	public static final String savedDisplayFilePfx = "FUN_SD_"; //$NON-NLS-1$
	public static final String savedDisplaySfx = ".txt"; //$NON-NLS-1$
	
	private ArrayList<OpenItem> openItems = new ArrayList<OpenItem>();
	private String focusedItem = "";
	
	private ArrayList<ConnectionProfile> connectionProfiles = new ArrayList<ConnectionProfile>();
	private boolean firstConnectionProfileLoad = true;
	
	private ArrayList<PreferenceCategory> preferenceCategories = new ArrayList<PreferenceCategory>();
	
	private boolean tracingActive = false;
	private FileWriter traceWtr = null;
	
	private ArrayList<DisplayListDeserializer> displayListDeserializers = new ArrayList<DisplayListDeserializer>();
	
	private ISavedDisplayManager savedDisplayManager = null;
	
	private boolean defaultConnectionProfileCreated = false;
	
	private Object lockEncrypt = new Object();
	private Object lockDecrypt = new Object();
	
	boolean firstTimeGettingStatuses = true;
	private static boolean runningWithIBMExplorer = false;
	private IListLayoutHelper	listLayoutHelper = null;
	private Environment defaultEnvironment = new Environment();
	public static final String IDENTIFIER	= "AppInstance";
	
	/**
	 * Added to support identification of the MenuItem for a <i>new</i> find operation, though wider use of this enum is expected 
	 * to occur, potentially encouraged.
	 * 
	 * @author mararn01
	 *
	 */
	public enum Flag {
		NEW("new"),
		DEFAULT("default");
		
		private String stringValue;
		
		//forcing the provision of a string representation and overriding toString should allow the enumeration objects to 
		//be used where either Object <i>or</i> String keys are required
		private Flag(String stringValue) {	
			this.stringValue = stringValue;
		}
		
		public String toString() {
			return stringValue;
		}
	}

	/**
	 * Added to support differentiation of the different NavigationProvider method calls while allowing the processing to be 
	 * performed in the one getMenuItems(...) method. 
	 * 
	 * It is expected that these values will become redundant if the needs of Context, Filter and Find methods diverge 
	 * significantly.
	 * 
	 * @author mararn01
	 *
	 */
	public enum Menu {
		CONTEXT,
		FILTER,
		FIND,
		EDIT;
	}
	
	public AppInstance() {
		
		super();
	}
	
	public AppInstance(BundleContext context, String userFolder, String type) {
		
		super();
		
		initialise(context.getDataFile("").getAbsolutePath(), userFolder, type); //$NON-NLS-1$
	}
	
	public AppInstance(String workingFolder, String userFolder, String type) {
		
		super();
		
		initialise(workingFolder, userFolder, type);
	}
	
	private void initialise(String workingFolder, String userFolder, String type) {
		
		instanceType = type;
		initialiseCipher();
		setWorkingFolder(workingFolder, userFolder);
		getPriorInstance();
		firstConnectionProfileLoad = true;
		retrieveConnectionProfiles();
		refreshNavigationProviders();
		listLayoutHelper = new ListLayoutHelper(this);
		loadListLayouts();
		obtainDisplayListDeserializers();
		obtainSavedDisplayManager();
		defaultEnvironment.put(IDENTIFIER, this);
	}

	private void initialiseCipher() {
	
		byte[] sd = new byte[] {67, 120, 78, 100, 84, 75, 54, 97, 81, 73, 101, 107, 85, 55, 49, 	53, 117, 80, 79, 	105, 118, 71, 102, 57};  //UTF-8 bytes.
		//Initialise a cipher suite to use for any encoding/decoding.
		try {
			SecureRandom sr = new SecureRandom(sd); //$NON-NLS-1$ //$NON-NLS-2$
			keyGenerator = KeyGenerator.getInstance(cipherAlgorithm);
			keyGenerator.init(168, sr);
			secretKey = keyGenerator.generateKey();
			cipher = Cipher.getInstance(cipherAlgorithm);
		}
		catch (Throwable e) {
			addConsoleMessage(e);
			secretKey = null;
		}
	}
	
	public void setWorkingFolder(String path, String userFolder) {
		
		String workPath;
		
		workingFolder = null;
		
		if (path == null) {
			workPath = ""; //$NON-NLS-1$
		}
		else {
			workPath = path;
		}
		if (userFolder != null) {
			if (userFolder.length() > 0) {
				workPath = workPath + File.separatorChar + userFolder;
				//Simulate the Eclipse workspace plug-in sub-folders so that the folder can be shared between Eclipse and Dojo.
				if (instanceType == instanceTypeDojo) {
					workPath = workPath + File.separatorChar + Messages.getString("AppInstance_23") + File.separatorChar + Messages.getString("AppInstance_24") + //$NON-NLS-1$ //$NON-NLS-2$
										File.separatorChar + Messages.getString("AppInstance_2"); //$NON-NLS-1$
				}
			}
		}
		
		if (workPath.length() > 0) {
			workingFolder = new File(workPath);
			if (!(workingFolder.exists())) {
				if (!(workingFolder.mkdirs())) {
					workingFolder = null;
				}
			}
			else {
				if (!(workingFolder.isDirectory())) {
					workingFolder = null;
				}
			}
		}
	}
	
	public String encrypt(String origValue) {
		
		synchronized (lockEncrypt) {
			String encryptedValue;
			BASE64Encoder encoder = new sun.misc.BASE64Encoder();

			try {
				cipher.init(Cipher.ENCRYPT_MODE, secretKey);
				byte[] origBytes = origValue.getBytes();
				byte[] encryptedBytes = cipher.doFinal(origBytes);
				encryptedValue = new String(encoder.encode(encryptedBytes));

				return encryptedValue;
			}
			catch (Exception e) {
				addConsoleMessage(e);
				return origValue;
			}
		}
	}
	
	public String decrypt(String encryptedValue) {
		
		synchronized (lockDecrypt) {
			
			String decryptedValue;
			BASE64Decoder decoder = new sun.misc.BASE64Decoder();
		
			try {
				cipher.init(Cipher.DECRYPT_MODE, secretKey);
				byte[] encryptedBytes = decoder.decodeBuffer(encryptedValue);
				byte[] decryptedBytes = cipher.doFinal(encryptedBytes);
				decryptedValue = new String(decryptedBytes);
			
				return decryptedValue;
			}
			catch (Exception e) {
				addConsoleMessage(e);
				return encryptedValue;
			}
		}
	}
	
	public String setConnectionDetails(String userid, String pswd, String newPswd, String timezone, boolean testConnections,
																String interval) {
		
		return  setConnectionDetails(userid, pswd, newPswd, timezone, testConnections, interval, true, false);
	}
	
	public String setConnectionDetails(String userid, String pswd, String newPswd, String timezone, boolean testConnections,
			String interval, boolean restoreEditors, boolean fromLogin) {

		int idxUsr = 0;
		ArrayList<AppUser> users = instance.getKnownUsers();
		AppUser user = null;
		boolean userKnown = false;
		boolean passwordCorrect = false;
		String testUserid = userid.trim();
		String testPswd = pswd.trim();
		String testNewPswd = newPswd.trim();
		String msg = messageOK;
		String testTimezone = timezone.trim();
		String testInterval = interval.trim();
		NumberEditor edt;
		
		if (instance.getCurrentUserid().length() > 0) {
			for (idxUsr = 0; idxUsr < users.size(); idxUsr++) {
				user = users.get(idxUsr);
				if (user.getUserid().equalsIgnoreCase(testUserid)) {
					userKnown = true;
					if (decrypt(user.getPassword()).equals(testPswd)) {
						passwordCorrect = true;
					}
					break;
				}
			}
		}
		else {
			if (testUserid.length() == 0) {
				userKnown = false;
			}
		}
		
		if (userKnown) {
			if (testPswd.length() > 0) {
				if (!passwordCorrect) {
					msg = messageInvalidUserPasswordCombination;
				}
			}
			else {
				if (testNewPswd.length() > 0) {
					msg = messageOldPasswordRequired;
				}
				else {
					if (fromLogin || !(testUserid.equals(instance.getCurrentUserid()))) {
						msg = messageNoPassword;
					}
				}
			}
		}
		else {
			if (testUserid.length() == 0) {
				msg = messageNoUser;
			}
			else {
				if (testPswd.length() == 0) {
					msg = messageNoPassword;
				}
			}
		}
		
		//If it's an Eclipse environment, ignore any userid/password related error messages, because we are
		//using connection profiles without any need for a userid or password.
		if ((!msg.equals(messageOK)) && instanceType.equals(instanceTypeEclipse)) {
			msg = messageOK;
		}
		
		if (msg.equals(messageOK)) {
			if (testTimezone.length() == 0) {
				msg = messageNoTimezone;
			}
		}
		
		if (msg.equals(messageOK)) {
			if (testInterval.length() == 0) {
				testInterval = Integer.toString(defaultSystemsRefreshInterval);
			}
			else {
				edt = new NumberEditor(testInterval);
				if (!(edt.isValidInteger(4))) {
					return messageInvalidTestInterval;
				}
				else {
					if (edt.intValue() <= 0) {
						return messageInvalidTestInterval;
					}
				}
			}
		}
		
		if (msg.equals(messageOK)) {
			instance.setCurrentUserid(testUserid);
			if (userKnown) {
				if (testNewPswd.length() > 0) {
					user.setPassword(encrypt(testNewPswd));
					instance.getKnownUsers().set(idxUsr, user);
				}
			}
			else {
				if (testPswd.length() > 0) {
					//Add the user to the list of known users, and set the current user to be them.
					user = new AppUser(testUserid, encrypt(testPswd));
					instance.addUser(user);
					instance.setCurrentUserid(user.getUserid());
				}
			}
			instance.setTimezone(testTimezone);
			instance.setRegularlyTestingConnections(testConnections);
			instance.setConnectionsTestInterval(Integer.parseInt(testInterval));
			instance.setRestoreEditorsAtStartup(restoreEditors);
			
			msg = saveInstance();
			
			manageSystemRefresh();
		}
		
		return msg;
	}
	
	public String setListPreferences(boolean keepKeysAtFront, ListFieldColor keyBackground, boolean saveListLayouts) {

		instance.setKeepListKeysAtFront(keepKeysAtFront);
		instance.setListKeyBackground(keyBackground);
		instance.setSavingListLayouts(saveListLayouts);

		return saveInstance();
	}
	
	public String setComparatorPreferences(ListFieldColor diffBackground, ListFieldColor sepBackground) {

		instance.setComparatorDiffBackground(diffBackground);
		instance.setComparatorSepBackground(sepBackground);

		return saveInstance();
	}
	
	public void setUnknownUser() {
		
		//Set the current user to an empty string to signify they're unknown (e.g. cancel from a login dialog).
		instance.setCurrentUserid(""); //$NON-NLS-1$
	}
	
	public String getInstanceType() {
		
		return instance.getInstanceType();
	}

	public void setInstanceType(String type) {
		
		instance.setInstanceType(type);
	}

	public int getTimeout() {
		
		return instance.getTimeout();
	}

	public String setTimeout(int timeoutSeconds) {
		
		instance.setTimeout(timeoutSeconds);
		return saveInstance();
	}

	public String getTitle() {
		
		return instance.getTitle();
	}

	public String setTitle(String title) {
		
		instance.setTitle(title);
		return saveInstance();
	}

	public String getCurrentUserid() { 
		
		return instance.getCurrentUserid();
	}
	
	public String getCurrentPassword() {
		
		return decrypt(instance.getCurrentPassword());
	}
	
	public String getTimezone() {
		
		return instance.getTimezone();
	}

	public String setTimezone(String timezone) {
		
		instance.setTimezone(timezone);
		return saveInstance();
	}

	public boolean regularlyTestingConnections() {
		
		return instance.regularlyTestingConnections();
	}

	public String setRegularlyTestingConnections(boolean testing) {
		
		instance.setRegularlyTestingConnections(testing);
		return saveInstance();
	}

	public int getConnectionsTestInterval() {
		
		return instance.getConnectionsTestInterval();
	}

	public String setConnectionsTestInterval(int intervalMinutes) {
		
		instance.setConnectionsTestInterval(intervalMinutes);
		return saveInstance();
	}

	public boolean restoreEditorsAtStartup() {
		
		return instance.restoreEditorsAtStartup();
	}

	public String setRestoreEditorsAtStartup(boolean restore) {
		
		instance.setRestoreEditorsAtStartup(restore);
		return saveInstance();
	}

	public boolean keepListKeysAtFront() {
		
		return instance.keepListKeysAtFront();
	}

	public boolean savingListLayouts() {
		
		return instance.savingListLayouts();
	}

	public String setKeepListKeysAtFront(boolean keepListKeysAtFront) {
		
		instance.setKeepListKeysAtFront(keepListKeysAtFront);
		return saveInstance();
	}

	public ListFieldColor getListKeyBackground() {
		
		return instance.getListKeyBackground();
	}

	public String setListKeyBackground(ListFieldColor listKeyBackground) {
		
		instance.setListKeyBackground(listKeyBackground);
		return saveInstance();
	}
	
	public ListFieldColor getComparatorDiffBackground() {
		
		return instance.getComparatorDiffBackground();
	}

	public String setComparatorDiffBackground(ListFieldColor compDiffBackground) {
		
		instance.setComparatorDiffBackground(compDiffBackground);
		return saveInstance();
	}
	
	public ListFieldColor getComparatorSepBackground() {
		
		return instance.getComparatorSepBackground();
	}

	public String setComparatorSepBackground(ListFieldColor compSepBackground) {
		
		instance.setComparatorSepBackground(compSepBackground);
		return saveInstance();
	}
	
	public boolean rememberConnectionSettings() {
		
		return instance.rememberConnectionSettings();
	}

	public String setRememberConnectionSettings(boolean remember) {
		
		instance.setRememberConnectionSettings(remember);
		return saveInstance();
	}
	
	public boolean showConnectionProfilesAtStartup() {
		 
		return instance.showConnectionProfilesAtStartup();
	}
	
	public String setShowConnectionProfilesAtStartup(boolean show) {
		
		instance.setShowConnectionProfilesAtStartup(show);
		return saveInstance();
	}
	
	public boolean tracingActive() {
		
		return tracingActive;
	}

	public String setTracingPreferences(boolean active, String path) {
		
		File file = null;
		boolean goodFile = false;
		
		tracingActive = active;
		instance.setTraceFilePath(path);
		
		if (active && path.length() <= 0) {
			return Messages.getString("AppInstance_28"); //$NON-NLS-1$
		}
		
		if (tracingActive) {
			file = new File(path);
			if (file.exists()) {
				if (file.isFile()) {
					if (file.canWrite()) {
						goodFile = true;
					}
				}
			}
			else {
				goodFile = true;
			}
			if (!goodFile) {
				return Messages.getString("AppInstance_3"); //$NON-NLS-1$
			}
				
			try {
				traceWtr = new FileWriter(file);
			}
			catch (IOException e) {
				addConsoleMessage(e);
				return Messages.getString("AppInstance_3"); //$NON-NLS-1$
			}
		}
		else {
			if (traceWtr != null) {
				try {
					traceWtr.flush();
					traceWtr.close();
				}
				catch (IOException e) {
					addConsoleMessage(e);
				}
				traceWtr = null;
			}
		}
		
		return saveInstance();
	}
	
	public void writeTrace(String issuer, String msg, String info) {
		
		String traceTime;
		String infoOut = ""; //$NON-NLS-1$
		
		if (tracingActive) {
			if (info != null) {
				infoOut = info; 
			}
			traceTime = Messages.getString("AppInstance_25") + (new SimpleDateFormat(Messages.getString("AppInstance_26"))).format(new Date()) + Messages.getString("AppInstance_27"); //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$
		
			try {
				traceWtr.write(traceTime + "\r\n" + issuer + ":  " + msg + "\r\n" + infoOut + "\r\n"); //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$ //$NON-NLS-4$
				traceWtr.flush();
			}
			catch (Throwable e) {
				addConsoleMessage(e);
			}
		}
	}
	
	public String getTraceFilePath() {
		
		return instance.getTraceFilePath();
	}
	
	public ArrayList<NavigationItem> getNavigationItems() {
	
		//This method should be called when the navigation item consumer listener is notified of a change to the
		//list of items.
		
		return allGroupedItems;
	}

	public void getNavigationItems(String filterType, String searchValue) {
		
		//This method should be called when the navigation item consumer wishes to change the criteria for
		//building the list of navigation items.
		
		this.filterType = filterType;
		this.searchValue = searchValue;
		rebuildNavigationItems();
	}
	
	public void rebuildNavigationInBackground() {
		
		RebuildNavigationThread thread;
		
		thread = new RebuildNavigationThread(this);
		thread.start();
		thread = null;
	}
	
	public synchronized void rebuildNavigationItems() {
		
		rebuildNavigationItems(true);
	}
	
	public synchronized void rebuildNavigationItems(boolean notifyListeners) {
		
		/*
		 * Get all the navigation items.
		 *
		 * Groups (including the system-built "All systems" group) are returned, and other items are
		 * returned as sub-items of those.
		 * 
		 * Connection Servers are only added to their own group.  Connection Servers and their sub-components (e.g. GPL repository) are not
		 * including in the main navigation tree.  We will probably remove them from the tree altogether at some later time, and maintain them
		 * elsewhere.
		 */
		
		Iterator<NavigationProvider> itProv;
		Iterator<NavigationItem> itItm;
		Iterator<NavigationItem> itGrp;
		Iterator<SerializableNavigationItem> itSave;
		ArrayList<NavigationItem> provItems = new ArrayList<NavigationItem>();
		ArrayList<SerializableNavigationItem> savedGroupItems = new ArrayList<SerializableNavigationItem>();
		NavigationItem allSystemsGroup = null;
		NavigationItem item;
		SerializableNavigationItem savedItem;
		NavigationItem group;
		NavigationProvider provider;
		boolean isUserGroup;
		Object data;
		boolean goodType;
		boolean satisfiesSearch;
		String srchval = searchValue.trim().toLowerCase();
		Iterator<NavigationRebuildListener> itListener;
		ArrayList<String> savedFullTitles = new ArrayList<String>();
		ArrayList<Integer> savedConnectStatuses = new ArrayList<Integer>();
		int idxSave;
		int idxItm;
		NavigationItemStatusesThread statusThread;
		ArrayList<NavigationItem> statusItems = new ArrayList<NavigationItem>();
		Method method = null;
		
		//Save currently known statuses, so they can be displayed until the statuses are refreshed.
		if (notifyListeners) {
			for (idxSave = 0; idxSave < connectionServerItems.size(); idxSave++) {
				item = connectionServerItems.get(idxSave);
				if (INavigationType.class.isInstance(item.getData())) {
					savedFullTitles.add(connectionServerItems.get(idxSave).getFullTitle());
					savedConnectStatuses.add(((INavigationType) item.getData()).getConnectStatus());
				}
			}
			for (idxSave = 0; idxSave < allItems.size(); idxSave++) {
				item = allItems.get(idxSave);
				if (INavigationType.class.isInstance(item.getData())) {
					savedFullTitles.add(allItems.get(idxSave).getFullTitle());
					savedConnectStatuses.add(((INavigationType) item.getData()).getConnectStatus());
				}
			}
		}
		
		getPriorInstance(false);  //This serves to clean/refresh the system definitions and navigation items by reloading them.
		
		allItems = null;
		allGroupedItems = null;
		allItems = new ArrayList<NavigationItem>();
		allGroupedItems = new ArrayList<NavigationItem>();
		viewableItems = new ArrayList<NavigationItem>();
		
		connectionServerItems.clear();
		for (itProv = navigationProviders.iterator(); itProv.hasNext();) {
			provider = itProv.next();
			provItems = provider.getNavigationItems(false);
			for (itItm = provItems.iterator(); itItm.hasNext();) {
				item = itItm.next();
				isUserGroup = false;
				item.setProvider(provider);
				if (item.getData() == null && item.getClass().getName().equals(NavigationItem.class.getName())) {
					isUserGroup = true;
					if (item.getDisplayCaption().equals(allSystemsGroupCaption)) {
						allSystemsGroup = item;
					}
				}
				if (isUserGroup) {
					allGroupedItems.add(item);
				}
				else {
					data = item.getData();
					
					//Connection Server?
					if (ConnectionServer.class.isInstance(data)) {
						goodType = false;
						connectionServerItems.add(item);
					}
					else {

						goodType = (filterType.length() == 0 || filterType.equals(allNavigationItemTypes));
						if (!goodType) {
							if (data != null) {
								if (getItemType(data).equals(filterType)) {
									goodType = true;
								}
							}
						}
						allItems.add(item);
						if (goodType) {
							satisfiesSearch = false;
							if (srchval.length() == 0) {
								satisfiesSearch = true;
							}
							else {
								if (item.getDisplayCaption().toLowerCase().contains(srchval)) {
									satisfiesSearch = true;
								}
							}
							if (satisfiesSearch) {
								viewableItems.add(item);
							}
						}
					}
				}
			}
		}
		
		/*
		 *  Do the building of sub-items.
		 *  
		 *  Note that for user-created groups, the sub-items are currently in the SerializedNavigationItem, because they were retrieved from
		 *  the saved Instance.  They just need to be added as "real" items to the group.
		 */
		for (itGrp = allGroupedItems.iterator(); itGrp.hasNext();) {
			group = itGrp.next();
			if (!(group.equals(allSystemsGroup))) {
				savedGroupItems = (ArrayList<SerializableNavigationItem>) group.getSerialized().getSubItems().clone();
				for (itSave = savedGroupItems.iterator(); itSave.hasNext();) {
					savedItem = itSave.next();
					for (itItm = viewableItems.iterator(); itItm.hasNext();) {
						item = itItm.next();
						if (item.getFullTitle().equals(savedItem.getFullTitle())) {
							addItemToNavigationGroup(group, item);
							break;
						}
					}
					for (itItm = allItems.iterator(); itItm.hasNext();) {
						item = itItm.next();
						if (item.getFullTitle().equals(savedItem.getFullTitle())) {
							addUnfilteredItemToNavigationGroup(group, item);
							break;
						}
					}
				}
			}
		}
		if (allSystemsGroup != null) {
			for (itItm = viewableItems.iterator(); itItm.hasNext();) {
				addItemToNavigationGroup(allSystemsGroup, itItm.next());
			}
			for (itItm = allItems.iterator(); itItm.hasNext();) {
				addUnfilteredItemToNavigationGroup(allSystemsGroup, itItm.next());
			}
		}
		
		for (itItm = connectionServerItems.iterator(); itItm.hasNext();) {
			allGroupedItems.add(itItm.next());
		}
		
		/*
		 *  Call any navigation providers who have an editNavigationItems() method, so they can manipulate the items if necessary.
		 *  
		 *  This can be used for adding extra data objects/merging items, where two plug-ins are providing essentially the same object
		 *  in different ways.
		 *  
		 *  This cannot be implemented via listeners because they run asynchronously and we need to be sure it is done before notifying
		 *  any navigation rebuild listeners.
		 *  
		 *  editNavigationItems() is not part of the NavigationProvider interface because it was implemented later, and we needed backwards
		 *  compatibility.
		 */
		for (itProv = navigationProviders.iterator(); itProv.hasNext();) {
			provider = itProv.next();
			method = null;
			try {
				method = provider.getClass().getMethod("editNavigationItems", new Class[] {allGroupedItems.getClass()}); //$NON-NLS-1$
			} 	
			catch (SecurityException e) {
					addConsoleMessage(e);
			} 	
			catch (NoSuchMethodException e) {}
			if (method != null) {
				try {
					method.invoke(provider, allGroupedItems);
					method.invoke(provider, allItems);
				} 				
				catch (Throwable e) {
					addConsoleMessage(e);			
				}
			}
		}
		
		/*
		 *  Set statuses to what we knew them as, notify listeners, then obtain fresh statuses (which notifies the listeners again to refresh the statuses).
		 *  
		 *  Refreshing the statuses is done in a background thread as it is expected to take a while if a lot of systems are defined.
		 */
		if (notifyListeners) {
			for (idxSave = 0; idxSave < savedFullTitles.size(); idxSave++) {
				for (idxItm = 0; idxItm < connectionServerItems.size(); idxItm++) {
					item = connectionServerItems.get(idxItm);
					if (INavigationType.class.isInstance(item.getData())) {
						if (item.getFullTitle().equals(savedFullTitles.get(idxSave))) {
							((INavigationType) item.getData()).setConnectStatus(savedConnectStatuses.get(idxSave));
						}
					}
				}
			}
			for (idxSave = 0; idxSave < savedFullTitles.size(); idxSave++) {
				for (idxItm = 0; idxItm < allItems.size(); idxItm++) {
					item = allItems.get(idxItm);
					if (INavigationType.class.isInstance(item.getData())) {
						if (item.getFullTitle().equals(savedFullTitles.get(idxSave))) {
							((INavigationType) item.getData()).setConnectStatus(savedConnectStatuses.get(idxSave));
						}
					}
				}
			}
			/*
			 *  Removed during z/OS Explorer integration, as it caused excessive flickering of the navigation view when z/OS Explorer API
			 *  triggered navigation rebuilds.  Kept commented out in case we want to know later where we had this.
			 *  
			 *  Since this change the statuses are all retrieved before we refresh the navigation.
			 *  
			 *  Since FUD-549 the following notify code was put back, because it caused a delay while statuses were re-obtained in the
			 *  background thread.  Needs to be revisited if we decide to reintroduce z/OS Explorer connection integration.
			 */
			if (!firstTimeGettingStatuses) {
				notifyNavigationListeners(allGroupedItems);
			}
			
			statusItems.addAll(allItems);
			statusItems.addAll(connectionServerItems);
			statusThread = new NavigationItemStatusesThread(this, statusItems, allGroupedItems);  
			statusThread.start();
			statusThread = null;
		}
	}
	
	public void notifyNavigationListeners(ArrayList<NavigationItem> items) {
		
		Iterator<NavigationRebuildListener> itListener;
		
		for (itListener = navigationRebuildListeners.iterator(); itListener.hasNext();) {
			itListener.next().navigationItemsChanged(items);
		}
	}
	
	public void notifyNavigationFocusListeners(Object source) {
		
		Iterator<NavigationFocusListener> itListener;
		
		for (itListener = navigationFocusListeners.iterator(); itListener.hasNext();) {
			if (NavigationItem.class.isInstance(source) && ((NavigationItem) source).getData() != null) {
				itListener.next().setFocus(((NavigationItem) source).getData());
			}
			else {
				itListener.next().setFocus(source);
			}
		}
	}
	
	public void notifyConnectionProfileEditListeners(ConnectionProfile profile) {
		
		Iterator<ConnectionProfileValidationListener> itListener;
		
		for (itListener = connectionProfileEditListeners.iterator(); itListener.hasNext();) {
			itListener.next().validateConnectionProfile(this, profile);
		}
	}
	
	private void addItemToNavigationGroup(NavigationItem group, NavigationItem item) {
		
		String type;
		boolean typeFnd;
		Iterator<NavigationItem> itType;
		NavigationItem typeItem = null;
		
		//If the navigation items are being filtered by type, add the item to the group because there is no type sub-item in each group. 
		//If not filtering by type, add the type first if necessary, then add the item to the type sub-item.
		if (!(filterType.length() == 0 || filterType.equals(allNavigationItemTypes))) {
			group.addSubItem(item);
		}
		else {
			type = getItemType(item.getData());
			if (type.length() > 0) {
				typeFnd = false;
				for (itType = group.getSubItems().iterator(); itType.hasNext();) {
					typeItem = itType.next();
					if (typeItem.getData() != null) {
						if (typeItem.getDisplayCaption().equals(type) && String.class.isInstance(typeItem.getData())) {
							typeFnd = true;
							break;
						}
					}
				}
				if (!typeFnd) {
					typeItem = new NavigationItem(type, type, null);
				}
				typeItem.addSubItem(item);
				group.addSubItem(typeItem);
			}
			else {
				group.addSubItem(item);
			}
		}
	}
	
	public ArrayList<NavigationItem> getUnfilteredGroupItems(String groupFullTitle) {
		
		int idx;
		ArrayList<NavigationItem> subItems = new ArrayList<NavigationItem>();
		NavigationItem group;
		
		for (idx = 0; idx < allGroupedItems.size(); idx++) {
			group = allGroupedItems.get(idx);
			if (group.getFullTitle().equals(groupFullTitle)) {
				subItems = (ArrayList<NavigationItem>) group.getUnfilteredSubItems(true).clone();
			}
		}
		
		return subItems;
	}
	
	private void addUnfilteredItemToNavigationGroup(NavigationItem group, NavigationItem item) {
		
		group.addUnfilteredSubItem(item);
	}
	
	public String getItemType(Object data) {
		
		String type = ""; //$NON-NLS-1$
		Method method = null;
		
		if (data != null) {
			try {
				method = data.getClass().getMethod(Messages.getString("AppInstance_30"), new Class[] {}); //$NON-NLS-1$
			} 	
			catch (SecurityException e) {
				addConsoleMessage(e);
			} 	
			catch (NoSuchMethodException e) {}
			if (method != null) {
				try {
					type = (String) method.invoke(data, new Object[] {});
				} 				
				catch (Throwable e) {
					addConsoleMessage(e);			
				}
			}
		}
		
		return type;
	}
	
	public ArrayList<String> getAllNavigationItemTypes() {
		
		ArrayList<String> types = new ArrayList<String>();
		Iterator<NavigationItem> itItm;
		Object data;
		Method method = null;
		String type;
		
		for (itItm = allItems.iterator(); itItm.hasNext();) {
			data = itItm.next().getData();
			//Don't add Connection Servers.
			if (!(data.getClass().isInstance(ConnectionServer.class))) {
				type = ""; //$NON-NLS-1$
				if (data != null) {
					try {
						method = data.getClass().getMethod(Messages.getString("AppInstance_32"), new Class[] {}); //$NON-NLS-1$
					} 	
					catch (SecurityException e) {
						addConsoleMessage(e);
					} 
					catch (NoSuchMethodException e) {}
					try {
						type = (String) method.invoke(data, new Object[] {});
					} 
					catch (Throwable e) {
						addConsoleMessage(e);
					}
				
					if (type.length() > 0) {
						if (!(types.contains(type))) {
							types.add(type);
						}
					}
				}
			}
		}
		
		Collections.sort(types);
		
		return types;
	}

	private void obtainDisplayListDeserializers() {
		
		IConfigurationElement[] extensions = null;
		IConfigurationElement element;
		int cnt;
		DisplayListDeserializer deserializer;
		
		extensions = refreshExtensions(Messages.getString("AppInstance_70")); //$NON-NLS-1$
		for (cnt = 0; cnt < extensions.length; cnt++) {
			element = extensions[cnt];
			try {
				deserializer = (DisplayListDeserializer) element.createExecutableExtension(Messages.getString("AppInstance_34")); //$NON-NLS-1$
				displayListDeserializers.add(deserializer);
			}
			catch (Throwable e) {
				addConsoleMessage(e);
			}
		}
	}
	
	private void obtainSavedDisplayManager() {
		
		IConfigurationElement[] extensions = null;
		IConfigurationElement element;
		
		//There should only ever be one - supplied by the Framework.
		extensions = refreshExtensions("com.fundi.framework.common.savedDisplayManager"); //$NON-NLS-1$
		if (extensions.length > 0) {
			element = extensions[0];
			try {
				savedDisplayManager = (ISavedDisplayManager) element.createExecutableExtension(Messages.getString("AppInstance_34")); //$NON-NLS-1$
				savedDisplayManager.setInstance(this);
			}
			catch (Throwable e) {
				addConsoleMessage(e);
			}
		}
	}
	
	public ArrayList<DisplayListDeserializer> getDisplayListDeserializers() {
		
		return displayListDeserializers;
	}

	public void refreshNavigationProviders() {
		
		IConfigurationElement[] extensions = null;
		IConfigurationElement element;
		int cnt;
		NavigationProvider navProvider;
		int iteration;
		
		extensions = refreshExtensions(Messages.getString("AppInstance_33")); //$NON-NLS-1$
		navigationProviders.clear();
		/*
		 * We need to ensure we get the FrameworkNavigationProvider (which should exist if this AppInstance class exists) first.
		 * This is because FrameworkNavigationProvider supplies the ConnectionServers, and they are needed before most other
		 * products (e.g. CEX, GPL) can be accessed.  It also supplies the generic main menu options.
		 */
		for (iteration = 0; iteration < 2; iteration++) {
			for (cnt = 0; cnt < extensions.length; cnt++) {
				element = extensions[cnt];
				try {
					navProvider = (NavigationProvider) element.createExecutableExtension(Messages.getString("AppInstance_34")); //$NON-NLS-1$
					if ((iteration == 0 && (FrameworkNavigationProvider.class.isInstance(navProvider))) ||
						(iteration > 0 && (!FrameworkNavigationProvider.class.isInstance(navProvider)))) {
						navProvider.initialise(this);
						navigationProviders.add(navProvider);
					}
				}
				catch (Throwable e) {
					addConsoleMessage(e);
				}
			}
		}
	}
		
	public IConfigurationElement[] refreshExtensions(String extensionPointId) {
		
		IExtensionRegistry registry;
		IConfigurationElement[] extensions = null;
		int cnt;
		
		registry = Platform.getExtensionRegistry();
		extensions = registry.getConfigurationElementsFor(extensionPointId);
		
		for (cnt = 0; cnt < extensions.length; cnt++) {
			reportLoadedBundle(extensions[cnt]);
		}
		
		return extensions;
	}
	
	private void reportLoadedBundle(IConfigurationElement element) {
		
		boolean alreadyLoaded = false;
		int idx;
		String name = ""; //$NON-NLS-1$
		String version = ""; //$NON-NLS-1$
		Dictionary<?, ?> bundleHeaders;
		
		bundleHeaders = Platform.getBundle(element.getDeclaringExtension().getContributor().getName()).getHeaders();
		name = (String) bundleHeaders.get(Messages.getString("AppInstance_37")); //$NON-NLS-1$
		version = (String) bundleHeaders.get(Messages.getString("AppInstance_38")); //$NON-NLS-1$
		
		for (idx = 0; idx < loadedBundleNames.size() && !alreadyLoaded; idx++) {
			if (loadedBundleNames.get(idx).equals(name) && loadedBundleVersions.get(idx).equals(version)) {
				alreadyLoaded = true;
			}
		}
		if (!alreadyLoaded) {
			loadedBundleNames.add(name);
			loadedBundleVersions.add(version);
			addConsoleMessage(new ConsoleMessage(name + Messages.getString("AppInstance_39") + version)); //$NON-NLS-1$
		}
	}
	
	public String getBundleResourceContent(String pluginName, String relativePath) {
		
		URL url;
		StringBuffer content = new StringBuffer();
		InputStream in;
		BufferedReader rdr;
		String line;
		
		url = Platform.getBundle(pluginName).getResource(relativePath);
		
		try {
			in = url.openConnection().getInputStream();
			rdr = new BufferedReader(new InputStreamReader(in));
		
			while ((line = rdr.readLine()) != null) {
				if (content.length() > 0) {
					content = content.append("\r\n"); //$NON-NLS-1$
				}
				content.append(line);
			}
		
			in.close();
		}
		catch (Throwable e) {
			addConsoleMessage(e);
		}
		
		return content.toString();
	}
	
	public ArrayList<NavigationItem> getNavigationGroups() {
		
		/*
		 * The "All systems" group is built at run time, and every system added to it.
		 */
		NavigationItem allSystemsGroup  = new NavigationItem(this);
		ArrayList<NavigationItem> groups = new ArrayList<NavigationItem>();
		ArrayList<SerializableNavigationItem> serGroups = instance.getNavigationGroups();
		Iterator<SerializableNavigationItem> itGrp;

		allSystemsGroup.setDisplayCaption(allSystemsGroupCaption);
		allSystemsGroup.setGroup(true);
		groups.add(0, allSystemsGroup);

		for (itGrp = serGroups.iterator(); itGrp.hasNext();) {
			groups.add(new NavigationItem(itGrp.next()));
		}
		
		return groups;
	}
	
	public ArrayList<NavigationType> getUserCreatableNavigationTypes() {
		
		Iterator<NavigationProvider> itProv;
		Iterator<NavigationType> itTyp;
		ArrayList<NavigationType> allTypes = new ArrayList<NavigationType>();
		ArrayList<NavigationType> provTypes = new ArrayList<NavigationType>();
		NavigationType type;
		NavigationProvider provider;
		
		for (itProv = navigationProviders.iterator(); itProv.hasNext();) {
			provider = itProv.next();
			provTypes = provider.getUserCreatableNavigationTypes();
			for (itTyp = provTypes.iterator(); itTyp.hasNext();) {
				type = itTyp.next();
				if (!(allTypes.contains(type))) {
					allTypes.add(type);
				}
			}
		}
		
		return allTypes;
	}
	
	public Object createNavigationItem(NavigationType type) {
		
		boolean refreshNavigation = false;
		NavigationItem ret;
		NavigationItemEditor editor = null;
		RebuildNavigationThread thread;
		
		if (type.equals(ConnectionServer.typeName) || type.equals(userGroupType)) {
			refreshNavigation = true;  //It's a Connection Server or group, so navigation probably changed.
		}
		
		editor = getNavigationItemEditor(type, null, true, false, false);
		if (editor != null) {
			ret = editor.create();
			
			//As of FUD-549, an Eclipse caller should initiate the refresh so it can use the progress view.  At the time of this change
			//only the NavigationView called this method, so following code commented out.
			
			//if (ret != null) {
			//	thread = new RebuildNavigationThread(this);
			//	thread.start();
			//	thread = null;
			//}
			
			return ret;
		}
		else {
			return null;
		}
	}
	
	public boolean deleteNavigationItem(NavigationItem item) {
		
		return deleteNavigationItem(item, true);
	}
	
	public boolean deleteNavigationItem(NavigationItem item, boolean rebuildNavigation) {
		
		boolean ret;
		NavigationItemEditor editor = null;
		RebuildNavigationThread thread;
		
		//As of 7/1/2013, always rebuild navigation, since if a plug-in is changing a navigation item, it probably affects the navigation.
		//As of 15/4/2013, this method is overloaded.  Eclipse-based NavigationItemEditors should initiate the rebuild of navigation, so that
		//their progress view can indicate it's happening.
		
		editor = getNavigationItemEditor(null, item, false, false, true);
		if (editor != null) {
			ret = editor.delete(item);

			if (ret && rebuildNavigation) {
				thread = new RebuildNavigationThread(this);
				thread.start();
				thread = null;
			}
			
			return ret;
		}
		else {
			return false;
		}
	}
	
	public boolean editNavigationItem(NavigationItem item) {
		
		return editNavigationItem(item, true);
	}
	
	public boolean editNavigationItem(NavigationItem item, boolean rebuildNavigation) {
		
		boolean ret;
		NavigationItemEditor editor = null;
		RebuildNavigationThread thread;
		
		//As of 7/1/2013, always rebuild navigation, since if a plug-in is changing a navigation item, it probably affects the navigation.
		//As of 15/4/2013, this method is overloaded.  Eclipse-based NavigationItemEditors should initiate the rebuild of navigation, so that
		//their progress view can indicate it's happening.
		
		editor = getNavigationItemEditor(null, item, false, true, false);
		if (editor != null) {
			ret = editor.edit(item);
			if (ret && rebuildNavigation) {
				thread = new RebuildNavigationThread(this);
				thread.start();
				thread = null;
			}
			
			return ret;
		}
		else {
			return false;
		}
	}
	
	public NavigationItemEditor getNavigationItemEditor(NavigationType type, NavigationItem item) {
		
		//Use this method to see if there is an editor available regardless of create/change/delete.
		return getNavigationItemEditor(type, item, true, true, true, null);
	}
	
	public NavigationItemEditor getNavigationItemEditor(NavigationType type, NavigationItem item, Environment environment) {
		
		//Use this method to see if there is an editor available regardless of create/change/delete.
		return getNavigationItemEditor(type, item, true, true, true, environment);
	}
	
	public NavigationItemEditor getNavigationItemEditor(NavigationType type, NavigationItem item, boolean create, boolean change, boolean delete) {

		return getNavigationItemEditor(type, item, true, true, true, null);
	}
	
	public NavigationItemEditor getNavigationItemEditor(NavigationType type, NavigationItem item, boolean create, boolean change, boolean delete, Environment environment) {
		
		IConfigurationElement[] extensions = null;
		IConfigurationElement element;
		int cnt;
		NavigationItemEditor navTypeEditor;
		
		extensions = refreshExtensions(Messages.getString("AppInstance_40")); //$NON-NLS-1$
		for (cnt = 0; cnt < extensions.length; cnt++) {
			element = extensions[cnt];
			try {
				navTypeEditor = (NavigationItemEditor) element.createExecutableExtension(Messages.getString("AppInstance_41")); //$NON-NLS-1$
				navTypeEditor.initialise(this, type, item, environment);
				if ((create && navTypeEditor.canCreateType()) || (change && navTypeEditor.canChangeType()) || (delete && navTypeEditor.canDeleteType())) {
					return navTypeEditor;
				}
			}
			catch (Throwable e) {
				addConsoleMessage(e);
			}
		}
		
		return null;
	}
	
	public ArrayList<MenuItem> getNavigationMenuItems(Environment environment) {
		
		/*
		 *   It's possible multiple extensions have created a menu item of the same caption.  Multiple non-parent duplicates are permitted,
		 *   and the extension developers need to sort out conflicts.
		 *   
		 *   In the duplicate parent menu case, use the first one.
		 */
		
		Iterator<NavigationProvider> itProv;
		Iterator<MenuItem> itItm;
		ArrayList<MenuItem> provItems = new ArrayList<MenuItem>();
		ArrayList<MenuItem> allItems = new ArrayList<MenuItem>();
		MenuItem item;
		NavigationProvider provider;
		
		for (itProv = navigationProviders.iterator(); itProv.hasNext();) {
			provider = itProv.next();
			provItems = provider.getNavigationMenuItems(environment);
			for (itItm = provItems.iterator(); itItm.hasNext();) {
				item = itItm.next();
				if (!(allItems.contains(item))) {
					allItems.add(item);
				}
			}
		}
		
		//If a within a category there are multiple parent menus of the same caption, merge them.
		allItems = mergeParentMenus(allItems);
		
		//Sort the menu items based on category, caption.
		sortMenuItems(allItems);
		
		//Add a separator between categories.
		allItems = addMenuSeparators(allItems);
		
		return allItems;
	}
	
	/**
	 * Convenience method.
	 * 
	 * @param environment	Map&lt;String, Object&gt; that allows arbitrary supporting objects to be passed - these are intended to be specific to an implementation environment such as a Viewer or a Request/Response
	 * @param sourceObject  This object represents the type of data that is being viewed, eg. a database etc.
	 * @return sorted list of MenuItem
	 */
	public ArrayList<MenuItem> getContextMenuItems(Environment environment, Object sourceObject) {
		return getMenuItems(Menu.CONTEXT, true, environment, sourceObject, null);
	}

	/**
	 * Convenience method.
	 * 
	 * @param environment	Map&lt;String, Object&gt; that allows arbitrary supporting objects to be passed - these are intended to be specific to an implementation environment such as a Viewer or a Request/Response
	 * @param sourceObject 	This object represents the type of data that is being viewed, eg. a database etc. 
	 * @param contextObject This object represents the user interface context of the view - eg. a list (editor), or a tree (navigation)
	 * @return sorted list of MenuItem
	 */
	public ArrayList<MenuItem> getContextMenuItems(Environment environment, Object sourceObject, Object contextObject) {
		return getMenuItems(Menu.CONTEXT, true, environment, sourceObject, contextObject);
	}

	/**
	 * Convenience method.
	 * 
	 * @param environment	Map&lt;String, Object&gt; that allows arbitrary supporting objects to be passed
	 * @param sourceObject	DisplayList the list to provide filtering MenuItems for.
	 * @return unsorted list of MenuItem
	 */
	public ArrayList<MenuItem> getFilterMenuItems(Environment environment, ListFilterHelper sourceObject) {
		return getMenuItems(Menu.FILTER, false, environment, sourceObject, null);
	}

	/**
	 * Convenience method.
	 * 
	 * @param environment	Map&lt;String, Object&gt; that allows arbitrary supporting objects to be passed
	 * @param sourceObject	List&lt;String&gt; historical list of keywords
	 * @return unsorted list of MenuItem
	 */
	public ArrayList<MenuItem> getFindMenuItems(Environment environment, List<DisplayList.FindItem> sourceObject) {
		return getMenuItems(Menu.FIND, false, environment, sourceObject, null);
	}

	/**
	 * Convenience method.
	 * 
	 * @param environment	Map&lt;String, Object&gt; that allows arbitrary supporting objects to be passed
	 * @param sourceObject	List&lt;String&gt; historical list of keywords
	 * @return unsorted list of MenuItem
	 */
	public ArrayList<MenuItem> getEditMenuItems(Environment environment, List<DisplayList.FindItem> sourceObject, IEditContext contextObject) {
		return getMenuItems(Menu.EDIT, false, environment, sourceObject, contextObject);
	}

	private ArrayList<MenuItem> getMenuItems(Menu menu, boolean sort, Environment environment, Object sourceObject, Object contextObject) {
		/*
		 *   It's possible multiple extensions have created a menu item of the same caption.  Multiple non-parent duplicates are permitted,
		 *   and the extension developers need to sort out conflicts.
		 *   
		 *   In the duplicate parent menu case, use the first one.
		 */
		
		Iterator<NavigationProvider> itProv;
		Iterator<MenuItem> itItm;
		ArrayList<MenuItem> provItems = new ArrayList<MenuItem>();
		ArrayList<MenuItem> allItems = new ArrayList<MenuItem>();
		MenuItem item;
		NavigationProvider provider;
		
		for (itProv = navigationProviders.iterator(); itProv.hasNext();) {
			provider = itProv.next();
			switch (menu) {
			case CONTEXT:
				//base NavigationProvider will call methods with the old signature, product plug-ins need to be mindful of this at the time that they override the new method signature
				provItems = provider.getContextMenuItems(environment, sourceObject, contextObject);
				break;

			case FILTER:
				//here sourceObject is really being used as a contextObject, might get a chance to change this one day
				provItems = provider.getFilterMenuItems(environment, sourceObject);	
				break;

			case FIND:
				//here sourceObject is really being used as a contextObject, might get a chance to change this one day
				provItems = provider.getFindMenuItems(environment, sourceObject);
				break;

			case EDIT:
				//here sourceObject is really being used as a contextObject, might get a chance to change this one day
				provItems = provider.getEditMenuItems(environment, sourceObject, contextObject);
				break;
	
			}
			
			for (itItm = provItems.iterator(); itItm.hasNext();) {
				item = itItm.next();
				if (!(allItems.contains(item))) {
					allItems.add(item);
				}
			}
		}
		
		//If a within a category there are multiple parent menus of the same caption, merge them.
		allItems = mergeParentMenus(allItems);
		
		if (sort) {
			//Sort the menu items based on category, caption.
			sortMenuItems(allItems);
		}
		
		//Add a separator between categories.
		allItems = addMenuSeparators(allItems);
		
		return allItems;
	}

	private ArrayList<MenuItem> mergeParentMenus(ArrayList<MenuItem> origItems) {
		
		Iterator<MenuItem> itItm;
		MenuItem item;
		ArrayList<MenuItem> mergedItems = new ArrayList<MenuItem>();
		int idxMerged;
		MenuItem mergeItem;
		boolean merged;	//mararn01: added this flag to ensure that an item with sub-items is added in the case that there is no existing item with the same caption (that has sub-items) for it to be merged into
		
		for (itItm = origItems.iterator(); itItm.hasNext();) {
			item = itItm.next();
			if (!item.hasSubItems()) {
				mergedItems.add(item);
			}
			else {
				merged = false;
				for (idxMerged = 0; idxMerged < mergedItems.size(); idxMerged++) {
					mergeItem = mergedItems.get(idxMerged);
					if (mergeItem.hasSubItems() && mergeItem.getCompareKey().equals(item.getCompareKey())) {
						mergeItem.addSubItems(mergeParentMenus(item.getSubItems()));  //Caution:  recursion.
						merged = true;
					}
				}
				if (merged == false) {
					mergedItems.add(item);
				}
			}
		}
		
		return mergedItems;
	}
	
	private void sortMenuItems(ArrayList<MenuItem> items) {
		
		Iterator<MenuItem> itItm;
		MenuItem parentItem;
		
		Collections.sort(items);
		
		for (itItm = items.iterator(); itItm.hasNext();) {
			parentItem = itItm.next();
			if (parentItem.hasSubItems()) {
				sortMenuItems(parentItem.getSubItems());  //Caution:  recursion.
			}
		}
	}
	
	private ArrayList<MenuItem> addMenuSeparators(ArrayList<MenuItem> origItems) {
		
		int idxItm;
		MenuItem item;
		ArrayList<MenuItem> separatedItems = new ArrayList<MenuItem>();
		
		for (idxItm = 0; idxItm < origItems.size(); idxItm++) {
			item = origItems.get(idxItm);
			if (item.hasSubItems()) {
				item.setSubItems(addMenuSeparators(item.getSubItems()));  //Caution:  recursion.
			}
			separatedItems.add(item);
			if (idxItm < (origItems.size() - 1)) {
				if (!(item.getCategory().equals(origItems.get(idxItm + 1).getCategory()))) {
					separatedItems.add(new MenuItem("", "", "", null, null, Style.SEPARATOR));  //Add a separator. //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$
				}
			}
		}
		
		return separatedItems;
	}
	
	public void openObjectEditors(Object sourceObject) {
		
		openObjectEditors(sourceObject, null, null);
	}
	
	public void openObjectEditors(Object sourceObject, Environment environment) {
		
		openObjectEditors(sourceObject, environment, null);
	}

	public void openObjectEditors(Object sourceObject, Environment environment, OpenItem priorItem) {
		
		IConfigurationElement[] extensions = null;
		IConfigurationElement element;
		int cnt;
		ObjectEditor objEditor;
		ArrayList<Object> allObjects = new ArrayList<Object>();
		NavigationItem item;
		ArrayList<Object> alternateObjects;
		int idxObj;
		
		/*
		 *  If sourceObject is a NavigationItem, it can have a data object and alternate data objects, e.g. where data is a CEX System
		 *  supplied by the CEX plug-in, and alternate data is a IMS Connect supplied by the GPL plug-in.  They logically refer to the same object
		 *  but are handled by different plug-ins.  Try to find an editor for each.
		 *  
		 *  If there is no data or alternate data then we're probably trying to open a NavigationItem, e.g. a group.  There can only be alternate data
		 *  if there is a data object.
		 */
		if (NavigationItem.class.isInstance(sourceObject)) {
			item = (NavigationItem) sourceObject;
			if (item.getData() != null) {
				allObjects.add(item.getData());
				alternateObjects = item.getAlternateData();
				for (idxObj = 0; idxObj < alternateObjects.size(); idxObj++) {
					allObjects.add(alternateObjects.get(idxObj));
				}
			}
			else {
				allObjects.add(sourceObject);
			}
		}
		else {
			allObjects.add(sourceObject);
		}
		
		
		extensions = refreshExtensions("com.fundi.framework.common.objectEditor"); //$NON-NLS-1$
		for (cnt = 0; cnt < extensions.length; cnt++) {
			element = extensions[cnt];
			try {
				objEditor = (ObjectEditor) element.createExecutableExtension("class"); //$NON-NLS-1$
				//objEditor.initialise(this, sourceObject, environment);
				//if (objEditor.canProvideEditor()) {
				//	objEditor.openEditor(priorItem);
				//}
				for (idxObj = 0; idxObj < allObjects.size(); idxObj++) {
					objEditor.initialise(this, allObjects.get(idxObj), environment);
					if (objEditor.canProvideEditor()) {
						objEditor.openEditor(priorItem);
					}
				}
			}
			catch (Throwable e) {
				addConsoleMessage(e);
			}
		}
	}
	
	public ArrayList<PreferenceCategory> getPreferenceCategoriesAndEditors(Environment environment) {
		
		IConfigurationElement[] extensions = null;
		IConfigurationElement element;
		int cnt;
		PreferenceEditor editor;

		preferenceCategories.clear();
		
		extensions = refreshExtensions("com.fundi.framework.common.preferenceEditor"); //$NON-NLS-1$
		for (cnt = 0; cnt < extensions.length; cnt++) {
			element = extensions[cnt];
			try {
				editor = (PreferenceEditor) element.createExecutableExtension("class"); //$NON-NLS-1$
				editor.initialise(this, environment);
				createPreferenceCategories(editor);
			}
			catch (Throwable e) {
				addConsoleMessage(e);
			}
		}
		
		return preferenceCategories;
	}
	
	private void createPreferenceCategories(PreferenceEditor editor) {
		
		ArrayList<PreferenceCategory> higherCategories;
		PreferenceCategory prefCat;
		int idx;
		PreferenceCategory cat;
		int idxCat;

		if (editor.getCategory() == null) {
			return;
		}
		
		prefCat = new PreferenceCategory(editor.getCategory(), editor);
		
		/*
		 *  The category at any level may have been created by an earlier preference editor provider.
		 *  If so, adjust its sub-category flag if necessary.
		 */
		higherCategories = prefCat.getHigherLevelCategories();
		for (idx = 0; idx < higherCategories.size(); idx++) {
			cat = higherCategories.get(idx);
			cat.setHasSubCategories(true);
			idxCat = preferenceCategories.indexOf(cat);
			if (idxCat < 0) {
				preferenceCategories.add(higherCategories.get(idx));
			}
			else {
				preferenceCategories.get(idxCat).setHasSubCategories(true);
			}
		}
		idxCat = preferenceCategories.indexOf(prefCat);
		if (idxCat < 0) {
			preferenceCategories.add(prefCat);
		}
		else {
			if (preferenceCategories.get(idxCat).getEditor() == null) {  //Don't override an already found editor - 1st gets used.
				preferenceCategories.get(idxCat).setEditor(prefCat.getEditor());
			}
		}
	}
	
	private void getPriorInstance() {
		
		getPriorInstance(true);
	}
	
	public void getPriorInstance(boolean manageRefresh) {
		
		File file;
		ObjectInputStream in;
		
		file = new File(workingFolder, instanceFilename);
		try {
			in = new ObjectInputStream(new FileInputStream(file));
			instance = (SerializableAppInstance) in.readObject();
		}
		catch (FileNotFoundException efnf) {
			instance = new SerializableAppInstance();
			instance.setInstanceType(instanceType);
		}
		catch (Throwable e) {
			instance = new SerializableAppInstance();
			addConsoleMessage(e);
		}

		if (manageRefresh) {
			manageSystemRefresh();
		}
	}
	
	private String saveInstance() {
		
		File saveFile;
		ObjectOutputStream out;
		String msg = ""; //$NON-NLS-1$
		
		if (workingFolder != null) {
			saveFile = new File (workingFolder, instanceFilename); 
			try {
				out = new ObjectOutputStream(new FileOutputStream(saveFile));
				out.writeObject(instance);
				out.flush();
				out.close();
			}
			catch (Throwable e) {
				addConsoleMessage(e);
				msg = exceptionToString(e);
			}
		}
		
		return msg;
	}
	
	public String createNavigationGroup(String name) {
		
		String msg = ""; //$NON-NLS-1$
		NavigationItem newGroup = null;
		String groupName = name.trim();
		ArrayList<SerializableNavigationItem> groups = instance.getNavigationGroups();
		
		Iterator<SerializableNavigationItem> itGrp;
		
		if (groupName.length() == 0) {
			msg = Messages.getString("AppInstance_49"); //$NON-NLS-1$
		}
		else {
			for (itGrp = groups.iterator(); itGrp.hasNext();) {
				if (itGrp.next().getDisplayCaption().equals(name)) {
					msg = Messages.getString("AppInstance_50"); //$NON-NLS-1$
					break;
				}
			}
			if (msg.length() == 0) {
				if (name.equalsIgnoreCase(AppInstance.allSystemsGroupCaption)) {
					msg = AppInstance.allSystemsGroupCaption + Messages.getString("AppInstance_51"); //$NON-NLS-1$
				}
			}
			
			if (msg.length() == 0) {
				newGroup = new NavigationItem(this);
				newGroup.setDisplayCaption(name);
				newGroup.setGroup(true);
				instance.addNavigationGroup(newGroup.serialize());
				msg = saveInstance();
			}
		}
		
		return msg;
	}
	
	public String deleteNavigationGroup(NavigationItem group) {
		
		String msg = ""; //$NON-NLS-1$
		ArrayList<SerializableNavigationItem> groups = instance.getNavigationGroups();
		Iterator<SerializableNavigationItem> itGrp;
		
		for (itGrp = groups.iterator(); itGrp.hasNext();) {
			if (itGrp.next().getDisplayCaption().equals(group.getDisplayCaption())) {
				itGrp.remove();
				msg = saveInstance();
				break;
			}
		}
		
		return msg;
	}
	
	public ArrayList<File> getUserFiles(String prefix) {
		
		ArrayList<File> files = new ArrayList<File>();
		String names[];
		int idx;
		
		names = workingFolder.list();
		for (idx = 0; idx < names.length; idx++) {
			if (prefix.equals("") || names[idx].startsWith(prefix)) { //$NON-NLS-1$
				files.add(new File(workingFolder, names[idx]));
			}
		}
		
		return files;
	}

	public File getWorkingFolder() {
		
		return workingFolder;
	}
	
	public ArrayList<NavigationItem> getAllItems() {
		
		Collections.sort(allItems);
		
		return allItems;
	}
	
public String adjustGroupItems(NavigationItem group, String[] inItems) {
		
		String msg = ""; //$NON-NLS-1$
		int idx;
		Iterator<NavigationItem> itItm;
		NavigationItem item;
		NavigationItem grp = null;
		ArrayList<NavigationItem> groups = getNavigationGroups();
		
		for (idx = 0; idx < groups.size(); idx++) {
			if (groups.get(idx).getFullTitle().equals(group.getFullTitle())) {
				grp = groups.get(idx);
				break;
			}
		}
		
		if (grp != null) {
			grp.removeAllSubItems();
		
			for (itItm = allItems.iterator(); itItm.hasNext();) {
				item = itItm.next();
				for (idx = 0; idx < inItems.length; idx++) {
					if (inItems[idx].equals(item.getFullTitle())) {
						grp.addSubItem(item);
					}
				}
			}
		
			saveInstance();
		}
		
		return msg;
	}
	
	public void addConsoleMessage(Throwable exc) {
		
		addConsoleMessage(new ConsoleMessage(exc.getMessage()));
		exc.printStackTrace();
	}

	public synchronized void addConsoleMessage(ConsoleMessage msg) {
	
		Iterator<ConsoleListener> itListener;
		
		if (msg.getMessage() != null && !msg.getMessage().isEmpty()) {
			consoleMessages.add(msg);
		}

		for (itListener = consoleListeners.iterator(); itListener.hasNext();) {
			itListener.next().consoleMessageAdded(msg);
		}
	}
	
	public void clearConsoleMessages() {
		
		consoleMessages.clear(); //$NON-NLS-1$
	}
	
	public synchronized void addConsoleListener(ConsoleListener listener) {
		
		consoleListeners.add(listener);
		
		//It's a new listener - messages may have been put in the console before this listener was added, so force their display.
		if (consoleMessages.size() > 0) {
			listener.consoleMessageAdded(consoleMessages);
		}
	}
	
	public synchronized void removeConsoleListener(ConsoleListener listener) {
		
		consoleListeners.remove(listener);
	}
	
	public synchronized void addNavigationRebuildListener(NavigationRebuildListener listener) {
		
		navigationRebuildListeners.add(listener);
	}
	
	public synchronized void addNavigationFocusListener(NavigationFocusListener listener) {
		
		navigationFocusListeners.add(listener);
	}
	
	public synchronized void removeNavigationRebuildListener(NavigationRebuildListener listener) {
		
		navigationRebuildListeners.remove(listener);
	}
	
	public synchronized void removeNavigationFocusListener(NavigationFocusListener listener) {
		
		navigationFocusListeners.remove(listener);
	}
	
	public synchronized void addConnectionProfileEditListener(ConnectionProfileValidationListener listener) {
		
		connectionProfileEditListeners.add(listener);
	}
	
	public synchronized void removeConnectionProfileEditListener(ConnectionProfileValidationListener listener) {
		
		connectionProfileEditListeners.remove(listener);
	}
	
	public String exceptionToString(Throwable exc) {
		
		StringBuffer ret = new StringBuffer(""); //$NON-NLS-1$
		int idx;
		StackTraceElement[] elements = exc.getStackTrace();
		
		ret.append(exc.toString());
		ret.append("\r\n"); //$NON-NLS-1$
		for (idx = 0; idx < elements.length; idx++) {
			ret.append(elements[idx].toString());
			ret.append("\r\n"); //$NON-NLS-1$
		}
		
		return ret.toString();
	}
	
	public ArrayList<ConnectionServer> retrieveSavedConnectionServers(boolean testConnections) {
		
		ArrayList<File> files = getUserFiles(connectionServerFilenamePfx);
		ObjectInputStream in = null;
		Iterator<File> itFile;
		ConnectionServer srvr;
		IConfigurationElement[] extensions = null;
		IConfigurationElement element;
		IConnectionServerProvider connSrvrProvider;

		connectionServers.clear();

		if (useIBMExplorerConnectionDefns()) {
			//Since we can't have a dependency on the Framework Eclipse plug-in, this needs to be done as an extension.
			extensions = refreshExtensions(Messages.getString("AppInstance_43")); //$NON-NLS-1$
			//We really only expect a single IConnectionServerProvider - the one that gets the Connection Server definitions from IBM Explorer.
			if (extensions.length > 0) {
				element = extensions[0];
				try {
					connSrvrProvider = (IConnectionServerProvider) element.createExecutableExtension(Messages.getString("AppInstance_34")); //$NON-NLS-1$
					connectionServers = connSrvrProvider.getConnectionServers(this);
				}
				catch (Throwable e) {
					addConsoleMessage(e);
				}
			}
		}
		else {
			for (itFile = files.iterator(); itFile.hasNext();) {
				try {
					in = new ObjectInputStream(new FileInputStream(itFile.next()));
					srvr = (ConnectionServer) in.readObject();
					if (testConnections) {
						srvr.testConnection(this);
					}
					connectionServers.add(srvr);
					in.close();
				} 
				catch (Throwable e) {
					addConsoleMessage(e);
				} 
			}
		}
		
		return connectionServers;
	}
	
	
	public ArrayList<ConnectionServer> getConnectionServers(boolean testConnections) {
		
		if (connectionServers.size() == 0) {
			retrieveSavedConnectionServers(testConnections);
		}

		return connectionServers;
	}
	
	public ArrayList<NavigationItem> getConnectionServerItems() {
		
		return connectionServerItems;
	}
	
	public String createConnectionServer(String name, String host, String port) {
		
		return createConnectionServer(name, host, port, ""); //$NON-NLS-1$
	}
	
	public String createConnectionServer(String name, String host, String port, String connectionProfileName) {
		
		String msg = ""; //$NON-NLS-1$
		String srvrName = name.trim();
		String srvrHost = host.trim();
		String srvrPort = port.trim();
		ConnectionServer srvr;
		
		msg = validateConnectionServerDetails(srvrName, srvrHost, srvrPort, true);
			
		if (msg.length() == 0) {
			srvr = new ConnectionServer(srvrName, srvrHost, Integer.parseInt(srvrPort));
			srvr.setConnectionProfileName(connectionProfileName);
			msg = saveConnectionServer(srvr);
			connectionServers.add(srvr);
		}
		
		return msg;
	}
	
	public String editConnectionServer(ConnectionServer connSrvr) {
		
		return editConnectionServer(connSrvr, "", true); //$NON-NLS-1$
	}
	
	public String editConnectionServer(ConnectionServer connSrvr, String connectionProfileName) {
		
		return editConnectionServer(connSrvr, connectionProfileName, true);
	}
		
	public String editConnectionServer(ConnectionServer connSrvr, String connectionProfileName, boolean saveIfGood) {
		
		int idx;
		String msg;
		
		connSrvr.setConnectionProfileName(connectionProfileName);
		
		msg = validateConnectionServerDetails(connSrvr.getName(), connSrvr.getHost(), Integer.toString(connSrvr.getPort()), false);
		
		if (msg.length() == 0 && saveIfGood) {
			if (connectionServers.size() == 0) {
				//Could be zero if they aren't retrieved yet, or if there aren't any, but we don't know the difference so try to retrieve anyway.
				//This is just a precaution, because to edit they should have already been retrieved.
				getConnectionServers(false);
			}
			for (idx = 0; idx < connectionServers.size(); idx++) {	
				if (connectionServers.get(idx).getName().equals(connSrvr.getName())) {
					connectionServers.set(idx, connSrvr);
					msg = saveConnectionServer(connSrvr);
				}
			}
		}
		
		return msg;
	}
	
	private String validateConnectionServerDetails(String name, String host, String port, boolean creatingServer) {
		
		String msg = ""; //$NON-NLS-1$
		String srvrName = name.trim();
		String srvrHost = host.trim();
		String srvrPort = port.trim();
		ArrayList<ConnectionServer> servers = getConnectionServers(false);
		Iterator<ConnectionServer> itSvr;
		
		if (srvrName.length() == 0 || srvrHost.length() == 0 || srvrPort.length() == 0) {
			msg = Messages.getString("AppInstance_60"); //$NON-NLS-1$
		}
		else {
			NumberEditor edt = new NumberEditor(srvrPort);
			if (edt.isValidInteger(5)) {
				if (edt.intValue() <= 0 || edt.intValue() > 65535) {
					msg = Messages.getString("AppInstance_61"); //$NON-NLS-1$
				}
			}
			else {
				msg = Messages.getString("AppInstance_62"); //$NON-NLS-1$
			}
		}
		
		if (msg.length() == 0 && creatingServer) {
			if (!(new StringEditor(srvrName).isValidFilename())) {
				msg = Messages.getString("AppInstance_1") + StringEditor.invalidFilenameCharacters; //$NON-NLS-1$
			}
			else {
				for (itSvr = servers.iterator(); itSvr.hasNext();) {
					if (itSvr.next().getName().equals(srvrName)) {
						msg = Messages.getString("AppInstance_63"); //$NON-NLS-1$
						break;
					}
				}
			}
		}
		
		return msg;
	}
	
	public String saveConnectionServer(ConnectionServer connSrvr) {
		
		ObjectOutputStream out = null;
		File srvrFile;
		
		srvrFile = new File(getWorkingFolder(), connectionServerFilenamePfx + connSrvr.getName() + Messages.getString("AppInstance_64")); //$NON-NLS-1$
		try {
			out = new ObjectOutputStream(new FileOutputStream(srvrFile.getAbsolutePath()));
			out.writeObject(connSrvr);
			out.flush();
			out.close();
			return ""; //$NON-NLS-1$
		}
		catch (Throwable e) {
			return exceptionToString(e);
		}
	}
	
	public String deleteConnectionServer(ConnectionServer connSrvr) {
		
		try {
			connectionServers.remove(connSrvr);
			new File(getWorkingFolder(), connectionServerFilenamePfx + connSrvr.getName() + Messages.getString("AppInstance_66")).delete(); //$NON-NLS-1$
		}
		catch (Throwable e) {
			return exceptionToString(e);
		}
		
		return ""; //$NON-NLS-1$
	}
	
	public ConnectionServer getConnectionServer(String name, String host, int port) {
		
		Iterator<ConnectionServer> itSvr;
		ConnectionServer server = null;
		
		if (connectionServers.size() == 0) {
			//Could be zero if they aren't retrieved yet, or if there aren't any, but we don't know the difference so try to retrieve anyway.
			getConnectionServers(false);
		}
		for (itSvr = connectionServers.iterator(); itSvr.hasNext();) {
			server = itSvr.next();
			if (server.getName().equals(name) && server.getHost().equals(host) && server.getPort() == port) {
				return server;
			}
		}
		
		return null;
	}
	
	private void manageSystemRefresh() {

		if (this.regularlyTestingConnections()) {
			if (refreshScheduler == null || systemRefreshService.isTerminated()) {
				final Runnable refresher = new Runnable() {
					public void run() {
						rebuildNavigationItems();
					}
				};
				refreshScheduler = systemRefreshService.scheduleAtFixedRate(refresher, getConnectionsTestInterval(), getConnectionsTestInterval(), TimeUnit.MINUTES);
			}
		}
		else {
			if (refreshScheduler != null && !(systemRefreshService.isTerminated())) {
				refreshScheduler.cancel(true);
				systemRefreshService.shutdownNow();
				try {
					while (!systemRefreshService.awaitTermination(10, TimeUnit.SECONDS));
				}
				catch (InterruptedException e) {}
			}
		}
	}
	
	public synchronized int registerComparatorSource(ComparatorSource source) {
		
		//This method is only required for registering open sources.  Closed sources are provided dynamically by AppInstance, as required.
		
		source.setId(++highestComparatorSourceId);
		openComparatorSources.add(source);
		
		return highestComparatorSourceId;
	}
	
	public synchronized void deregisterComparatorSource(int id) {
	
		int idx;
		
		if (id >= 0) {
			for (idx = 0; idx < openComparatorSources.size(); idx++) {
				if (openComparatorSources.get(idx).getId() == id) {
					openComparatorSources.set(idx, null);
					openComparatorSources.remove(idx);
					return;
				}
			}
			for (idx = 0; idx < savedDisplayComparatorSources.size(); idx++) {
				if (savedDisplayComparatorSources.get(idx).getId() == id) {
					savedDisplayComparatorSources.set(idx, null);
					savedDisplayComparatorSources.remove(idx);
					return;
				}
			}
		}
	}
	
	public synchronized void deregisterComparatorSource(ComparatorSource source) {
		
		openComparatorSources.remove(source);
		source = null;
	}
	
	public synchronized void openComparator(int id) {
		
		int idx;
		ComparatorSource sce = null;
		
		if (id >= 0) {
			for (idx = 0; idx < openComparatorSources.size(); idx++) {
				if (openComparatorSources.get(idx).getId() == id) {
					sce = openComparatorSources.get(idx);
					break;
				}
			}
		}
		
		if (sce == null) {
			sce = new ComparatorSource("", null);
			sce.setId(-1);
		}

		if (sce != null) {
			openObjectEditors(sce);
		}
	}
	
	public synchronized ComparatorSource getComparatorSource(int id) {
		
		int idx;
		ComparatorSource sce = null;
		
		if (id >= 0) {
			for (idx = 0; idx < openComparatorSources.size(); idx++) {
				if (openComparatorSources.get(idx).getId() == id) {
					return openComparatorSources.get(idx);
				}
			}
			for (idx = 0; idx < savedDisplayComparatorSources.size(); idx++) {
				if (savedDisplayComparatorSources.get(idx).getId() == id) {
					return savedDisplayComparatorSources.get(idx);
				}
			}
		}
		
		return sce;
	}
	
	public synchronized ArrayList<ComparatorSource> getOpenComparatorSources() {
		
		return openComparatorSources;
	}
	
	public String writeSavedDisplay(Object display, Object titles, Object lists, File saveFile) {
		
		return writeSavedDisplay(null, display, titles, lists, saveFile);
	}
	
	public String writeSavedDisplay(String deserializerClass, Object display, Object titles, Object lists, File saveFile) {
		
		ObjectOutputStream out;
		String msg = ""; //$NON-NLS-1$
		
		try{
			out = new ObjectOutputStream(new FileOutputStream(saveFile));
			
			if (deserializerClass != null) {
				out.writeObject(Messages.getString("AppInstance_72") + deserializerClass); //$NON-NLS-1$
			}
			out.writeObject(titles);
			out.writeObject(lists);
			out.writeObject(display);
			out.flush();
			out.close();
		}
		catch (Throwable e) {
			addConsoleMessage(e);
			msg = exceptionToString(e);
		}
		
		return msg;
	}
	
	public synchronized ArrayList<ComparatorSource>getSavedDisplayComparatorSources() {
		
		return savedDisplayComparatorSources;
	}
	
	public synchronized void retrieveSavedDisplayComparatorSources() {
		
		ArrayList<File> files = getUserFiles(savedDisplayPfx);
		ObjectInputStream in = null;
		Iterator<File> itFile;
		File file;
		String title = ""; //$NON-NLS-1$
		Object titles = null;
		ArrayList<String> titlesList;
		ComparatorSource source = null;
		int idx;
		int idxDisp;
		ArrayList<SavedDisplay> savedDisplays = new ArrayList<SavedDisplay>();
		SavedDisplay saved;
	
		//Deregister any saved displays already retrieved, ready for a new list.
		for (idx = 0; idx < savedDisplayComparatorSources.size(); idx++) {
			deregisterComparatorSource(savedDisplayComparatorSources.get(idx));
		}
		
		savedDisplayComparatorSources.clear();
		
		//Get any new style saved displays.
		savedDisplays = savedDisplayManager.getAllSavedDisplays();
		for (idxDisp = 0; idxDisp < savedDisplays.size(); idxDisp++) {
			saved = savedDisplays.get(idxDisp);
			titlesList = saved.getComparatorSourceTitles();
			for (idx = 0; idx < titlesList.size(); idx++) {
				source = new ComparatorSource(titlesList.get(idx), saved);
				source.setId(++highestComparatorSourceId);
				savedDisplayComparatorSources.add(source);
			}
		}
		
		/*
		 *  We only retrieve the title for saved display sources, because opening each source to get the display list is too slow when
		 *  there are a number of saved display sources with large lists.
		 *  
		 *  A saved display may specify a deserializer class as the first object within it, used when the objects in list rows require a
		 *  non-common code class to be deserialized.
		 *  
		 *  The data object passed is the File object, and if this source is later selected then the Comparator will open that file later
		 *   to retrieve the display list.
		 *   
		 *   It is possible for a saved display comparator source to contain multiple lists, in which case it also has a corresponding array
		 *   of multiple titles (CEX plug-in does this for its session and status monitor editors).  In this case create multiple comparator sources,
		 *   one for each list.
		 */
		for (itFile = files.iterator(); itFile.hasNext();) {
			try {
				file = itFile.next();
				in = new ObjectInputStream(new FileInputStream(file));
				titles = in.readObject();
				if (String.class.isInstance(titles)) {
					if (((String) titles).toLowerCase().startsWith(Messages.getString("AppInstance_72"))) {  //Ignore it. //$NON-NLS-1$
						titles = in.readObject();
					}
				}
				if (ArrayList.class.isInstance(titles)) {
					titlesList = (ArrayList<String>) titles;
					for (idx = 0; idx < titlesList.size(); idx++) {
						source = new ComparatorSource(titlesList.get(idx), file);
						source.setId(++highestComparatorSourceId);
						savedDisplayComparatorSources.add(source);
					}
				}
				else {
					title = (String) titles;
					source = new ComparatorSource(title, file);
					source.setId(++highestComparatorSourceId);
					savedDisplayComparatorSources.add(source);
				}
				in.close();
			} 
			catch (Throwable e) {
				addConsoleMessage(e);
			} 
		}
	}
	
	public void setOpenItem(String fullTitle, String editorTitle, String address) {
		
		boolean fnd = false;
		int idx;
		OpenItem item;
		
		for (idx = 0; idx < openItems.size(); idx++) {
			item = openItems.get(idx);
			if (item.getSourceObjectFullTitle().equals(fullTitle)) {
				fnd = true;
				item.setEditor(editorTitle, address);
				break;
			}
		}
		
		if (!fnd) {
			item = new OpenItem(fullTitle);
			item.setEditor(editorTitle, address);
			openItems.add(item);
		}
	}
	
	public void removeOpenItem(String fullTitle, String editorTitle) {
		
		int idx;
		OpenItem item;
		
		for (idx = 0; idx < openItems.size(); idx++) {
			item = openItems.get(idx);
			if (item.getSourceObjectFullTitle().equals(fullTitle)) {
				item.removeEditor(editorTitle);
				if (item.getEditorCount() == 0) {
					openItems.remove(idx);
				}
			}
		}
	}
	
	public void saveOpenItems() {
		
		ObjectOutputStream out = null;
		File srvrFile;
		int idx;
		OpenItem item;
		
		//Save the focused one last, so that it is the last one to get loaded next time and hence retain the focus.
		for (idx = 0; idx < openItems.size(); idx++) {
			item = openItems.get(idx);
			if (item.getSourceObjectFullTitle().equals(focusedItem)) {
				openItems.remove(idx);
				openItems.add(item);  //Force it to the last one.
			}
		}
		
		srvrFile = new File(getWorkingFolder(), openItemsFilename);
		try {
			out = new ObjectOutputStream(new FileOutputStream(srvrFile.getAbsolutePath()));
			out.writeObject(openItems);
			out.flush();
			out.close();
		}
		catch (Throwable e) {
			addConsoleMessage(e);
		}
	}
	
	public void restorePriorOpenItems(Environment environment) {
		
		ObjectInputStream in = null;
		int idxItm;
		int idxView;
		ArrayList<File> savedCompares;
		int idxCmp;
	
		/*
		 *  Look for any prior comparator instances to open.  They were saved as saved displays because they can't be opened
		 *  from a navigation item.
		 *  
		 *  To restore them, we open a comparator for each one, and if it hasn't been passed a good comparator source to open then
		 *  it will look for a saved display to open.
		 *  
		 *  Do this before other editors, so that a comparator isn't the last one with focus.
		 */
		savedCompares = getUserFiles(savedDisplayFilePfx + ComparatorHelper.comparatorSavedDisplayPfx);
		for (idxCmp = 0; idxCmp < savedCompares.size(); idxCmp++) {
			openComparator(-1);  //Dummy comparator source id, as the editor will look for the saved display to open.
		}

		openItems.clear();
		try {
			in = new ObjectInputStream(new FileInputStream(new File(workingFolder, openItemsFilename)));
			openItems = (ArrayList<OpenItem>) in.readObject();
			in.close();
		} 
		catch (FileNotFoundException fne) {}
		catch (Throwable e) {
			addConsoleMessage(e);
		}
		
		for (idxItm = 0; idxItm < openItems.size(); idxItm++) {
			//Look up viewable items for non-group items to reopen.
			for (idxView = 0; idxView < viewableItems.size(); idxView++) {
				if (openItems.get(idxItm).getSourceObjectFullTitle().equals(viewableItems.get(idxView).getFullTitle())) {
					openObjectEditors(viewableItems.get(idxView), environment, openItems.get(idxItm));
				}
			}
			//Look up group items for groups to reopen.
			for (idxView = 0; idxView < allGroupedItems.size(); idxView++) {
				if (openItems.get(idxItm).getSourceObjectFullTitle().equals(allGroupedItems.get(idxView).getFullTitle())) {
					if (allGroupedItems.get(idxView).getData() != null) {
						openObjectEditors(allGroupedItems.get(idxView).getData(), environment, openItems.get(idxItm));
					}
					else {
						openObjectEditors(allGroupedItems.get(idxView), environment, openItems.get(idxItm));
					}
				}
			}
		}
	}
	
	public ArrayList<ConnectionProfile> getConnectionProfiles() {
		
		return connectionProfiles;
	}
	
	public ConnectionProfile getDefaultConnectionProfile() {
		
		int idx;
		ConnectionProfile profile = null;
		
		for (idx = 0; idx < connectionProfiles.size(); idx++) {
			profile = connectionProfiles.get(idx);
			if (profile.isDefaultProfile()) {
				break;
			}
		}
		
		return profile;
	}

	public ConnectionProfile getConnectionProfile(String connectionProfileName) {
		
		int idx;
		ConnectionProfile testProfile = null;
		ConnectionProfile profile = null;
		ConnectionProfile defaultProfile = null;
		
		for (idx = 0; idx < connectionProfiles.size(); idx++) {
			testProfile = connectionProfiles.get(idx);
			if (testProfile.getName().equals(connectionProfileName)) {
				profile = testProfile;
				break;
			}
			if (testProfile.isDefaultProfile()) {
				defaultProfile = testProfile;
			}
		}
		
		if (profile == null) {
			profile = defaultProfile;
			if (profile == null) {
				profile = new ConnectionProfile();
			}
		}
		
		return profile;
	}
	
	public String getConnectionUserid(String connectionProfileName) {
		
		return getConnectionProfile(connectionProfileName).getUserid();
	}
	
	public String getConnectionPassword(String connectionProfileName) {
		
		String password = getConnectionProfile(connectionProfileName).getPassword();
		
		if (password.length() > 0) {
			password = decrypt(password);
		}
		
		return password;
	}
	
	public String setConnectionProfiles(ArrayList<ConnectionProfile> profiles, boolean rememberConnectionSettings, boolean showAtStartup,
														boolean useIBMExplorer) {
		
		//if IBM Explorer is not being used to manage them, don't overwrite any connection profiles we have previously managed, in case
		//the user wants to go back to them.
		if (!useIBMExplorer) {
			connectionProfiles = profiles;
		}
		instance.setRememberConnectionSettings(rememberConnectionSettings);
		instance.setShowConnectionProfilesAtStartup(showAtStartup);
		instance.setUseIBMExplorerConnectionDefns(useIBMExplorer);
		
		return saveInstance();
	}

	public void setConnectionProfile(ConnectionProfile profile) {
		
		boolean fnd = false;
		int idx;
		ConnectionProfile connProf;
		
		for (idx = 0; idx < connectionProfiles.size(); idx++) {
			connProf = connectionProfiles.get(idx);
			if (connProf.getName().equals(profile.getName())) {
				fnd = true;
				connectionProfiles.set(idx, profile);
				break;
			}
		}
		
		if (!fnd) {
			connectionProfiles.add(profile);
		} 
		
		saveConnectionProfiles();
	}
	
	public void removeConnectionProfile(ConnectionProfile profile) {
		
		int idx;
		ConnectionProfile connProf;
		
		for (idx = 0; idx < connectionProfiles.size(); idx++) {
			connProf = connectionProfiles.get(idx);
			if (connProf.getName().equals(profile.getName())) {
				connectionProfiles.remove(idx);
				break;
			}
		}
		
		saveConnectionProfiles();
	}
	
	public String saveConnectionProfiles() {
		
		ObjectOutputStream out = null;
		File file;
		String msg = ""; //$NON-NLS-1$
		ArrayList<ConnectionProfile> saveProfiles = new ArrayList<ConnectionProfile>();
		int idx;
		ConnectionProfile profile;
		
		for (idx = 0; idx < connectionProfiles.size(); idx++) {
			profile = new ConnectionProfile(connectionProfiles.get(idx));
			if (!instance.rememberConnectionSettings()) {
				profile.setPassword(encrypt(""));
			}
			saveProfiles.add(profile);
		}
		
		file = new File(getWorkingFolder(), connectionProfilesFilename);
		try {
			out = new ObjectOutputStream(new FileOutputStream(file.getAbsolutePath()));
			out.writeObject(saveProfiles);
			out.flush();
			out.close();
		}
		catch (Throwable e) {
			addConsoleMessage(e);
			msg = e.getMessage();
		}
		
		return msg;
	}
	
	public void retrieveConnectionProfiles() {
		
		ObjectInputStream in = null;
		ConnectionProfile profile;
		boolean defaultExists = false;
		int idx;
		boolean profilesChanged = false;
		IConfigurationElement[] extensions = null;
		IConfigurationElement element;
		IConnectionServerProvider connSrvrProvider;
		connectionProfiles.clear();
		
		if (useIBMExplorerConnectionDefns()) {
			//Since we can't have a dependency on the Framework Eclipse plug-in, this needs to be done as an extension.
			extensions = refreshExtensions(Messages.getString("AppInstance_43")); //$NON-NLS-1$
			//We really only expect a single IConnectionServerProvider - the one that gets the Connection Server definitions from IBM Explorer.
			if (extensions.length > 0) {
				element = extensions[0];
				try {
					connSrvrProvider = (IConnectionServerProvider) element.createExecutableExtension(Messages.getString("AppInstance_34")); //$NON-NLS-1$
					connectionProfiles = connSrvrProvider.getConnectionProfiles(this);
				}
				catch (Throwable e) {
					addConsoleMessage(e);
				}
			}
		}
		else {
		
			try {
				in = new ObjectInputStream(new FileInputStream(new File(workingFolder, connectionProfilesFilename)));
				connectionProfiles = (ArrayList<ConnectionProfile>) in.readObject();
				in.close();
			} 
			catch (FileNotFoundException fne) {}
			catch (Throwable e) {
				addConsoleMessage(e);
				return;
			}
		
			//If there are no connection profiles, but we have a userid, create a default profile from it.
		
			//If the default connection profile doesn't yet exist, create it.  If it does, ensure it is set as default.
			for (idx = 0; idx < connectionProfiles.size(); idx++) {
				profile = connectionProfiles.get(idx);
				if (profile.getName().equals(defaultConnectionProfile)) {
					if (!(profile.isDefaultProfile())) {
						profile.setDefaultProfile(true);
						profilesChanged = true;
					}
					defaultExists = true;
				}
				else {
					if (profile.isDefaultProfile()) {
						profile.setDefaultProfile(false);
						profilesChanged = true;
					}
				}
			}
			if (!defaultExists) {
				profile = new ConnectionProfile();
				profile.setName(defaultConnectionProfile);
				profile.setDefaultProfile(true);

				if (instance.getCurrentUserid().length() > 0) {
					profile.setUserid(instance.getCurrentUserid());
					profile.setPassword(instance.getCurrentPassword());  //It's already encrypted, so no manipulation required.
				}
				else {
					profile.setUserid(Messages.getString("AppInstance_10")); //$NON-NLS-1$
					profile.setPassword(encrypt(Messages.getString("AppInstance_10"))); //$NON-NLS-1$
				}
				connectionProfiles.add(profile);
				profilesChanged = true;
				defaultConnectionProfileCreated = true;
			}
		
			if (profilesChanged) {
				saveConnectionProfiles();
			}
		
			//Ensure passwords are removed if not remembering them across sessions (in case previously they were saved to disk).
			if (!instance.rememberConnectionSettings()) {
				for (idx = 0; idx < connectionProfiles.size(); idx++) {
					connectionProfiles.get(idx).setPassword(encrypt(""));
				}
			}
		}
		
		if (firstConnectionProfileLoad) {
			//If the first time, reset the useability of each profile.
			for (idx = 0; idx < connectionProfiles.size(); idx++) {
				profile = connectionProfiles.get(idx);
				profile.setUseable(true);
			}
			firstConnectionProfileLoad = false;
		}
	}
	
	public ArrayList<ConnectionProfile> readSavedConnectionProfiles() {
		
		ObjectInputStream in = null;
		ArrayList<ConnectionProfile> profiles = new ArrayList<ConnectionProfile>();
		
		try {
			in = new ObjectInputStream(new FileInputStream(new File(workingFolder, connectionProfilesFilename)));
			profiles = (ArrayList<ConnectionProfile>) in.readObject();
			in.close();
		} 
		catch (FileNotFoundException fne) {}
		catch (Throwable e) {
			addConsoleMessage(e);
		}
		
		return profiles;
	}
	
	public void saveExpandedNavigationItems(ArrayList<String> items) {
		
		ObjectOutputStream out = null;
		File srvrFile;
		
		srvrFile = new File(getWorkingFolder(), expandedNavItemsFilename);
		try {
			out = new ObjectOutputStream(new FileOutputStream(srvrFile.getAbsolutePath()));
			out.writeObject(items);
			out.flush();
			out.close();
		}
		catch (Throwable e) {
			addConsoleMessage(e);
		}
	}
	
	public ArrayList<String> getPriorExpandedNavigationItems(Environment environment) {
		
		ObjectInputStream in = null;
		ArrayList<String> itemTitles = new ArrayList<String>();
		
		try {
			in = new ObjectInputStream(new FileInputStream(new File(workingFolder, expandedNavItemsFilename)));
			itemTitles = (ArrayList<String>) in.readObject();
			in.close();
		} 
		catch (FileNotFoundException fne) {}
		catch (Throwable e) {
			addConsoleMessage(e);
		}

		return itemTitles;
	}

	@Override
	protected void finalize() throws Throwable {
		
		super.finalize();
		
		if (traceWtr != null) {
			traceWtr.flush();
			traceWtr.close();
		}
	}

	public FileWriter getTraceWtr() {
		
		return traceWtr;
	}

	public SerializableAppInstance getInstance() {
		return instance;
	}
	
	public String getApplicationParm(String parmName, String defaultValue) {
		
		return instance.getApplicationParm(parmName, defaultValue);
	}
	
	public void setApplicationParm(String parmName, boolean value) {
		
		if (value) {
			setApplicationParm(parmName, "true"); //$NON-NLS-1$
		}
		else {
			setApplicationParm(parmName, "false"); //$NON-NLS-1$
		}
	}
	
	public void setApplicationParm(String parmName, String value) {
		
		String msg;
		
		instance.setApplicationParm(parmName, value);
		
		msg = saveInstance();
		if (msg.length() > 0) {
			addConsoleMessage(new ConsoleMessage(msg));
		}
	}
	
	public void setListLayout(DisplayList list) {
		listLayoutHelper.set(list);
	}
	
	public ListLayoutV1 getListLayout(String context) {
		return listLayoutHelper.get(context);
	}
	
	public void deleteListLayout(String context) {
		listLayoutHelper.delete(context);
	}
	
	public void saveListLayouts() {
		listLayoutHelper.saveAll();
	}
	
	public void loadListLayouts() {
		listLayoutHelper.loadAll();
	}
	
	public IListLayoutHelper getListLayoutHelper() {
		return listLayoutHelper;
	}

	public boolean defaultConnectionProfileCreated() {
		return defaultConnectionProfileCreated;
	}

	public boolean isFirstTimeGettingStatuses() {
		return firstTimeGettingStatuses;
	}

	public void setFirstTimeGettingStatuses(boolean firstTimeGettingStatuses) {
		this.firstTimeGettingStatuses = firstTimeGettingStatuses;
	}

	public static boolean isRunningWithIBMExplorer() {
		return runningWithIBMExplorer;
	}

	public static void setRunningWithIBMExplorer(boolean withIBMExplorer) {
		
		runningWithIBMExplorer = withIBMExplorer;
	}
	
	public boolean useIBMExplorerConnectionDefns() {
		
		return runningWithIBMExplorer && instance.useIBMExplorerConnectionDefns();
	}

	public ISavedDisplayManager getSavedDisplayManager() {
		
		return savedDisplayManager;
	}

	public void setFocusedItem(String focusedItem) {
		
		this.focusedItem = focusedItem;
	}
}
